import typing as t
from datetime import date, time, timedelta
from functools import partial

from click import Context, Parameter
from click.core import ParameterSource
from click.shell_completion import CompletionItem
from django.conf import settings
from django.db import models
from django.db.models.query import QuerySet


def int_ranges(incomplete: str, max_val: int) -> t.List[t.Tuple[int, int]]:
    lower = int(incomplete)
    neg = lower < 0
    lower = abs(lower)
    upper = abs(lower + 1)
    ranges = [(-upper, -lower)] if neg else [(lower, upper)]
    while (lower := lower * 10) <= max_val:
        upper *= 10
        ranges.append((-upper, -lower) if neg else (lower, upper))
    return ranges


def int_query(
    incomplete: str,
    lookup_field: str,
    queryset: QuerySet,
    **_,
) -> models.Q:
    """
    The default completion query builder for integer fields. This method will
    return a Q object that will match any value that starts with the incomplete
    string. For example, if the incomplete string is "1", the query will match
    1, 10-19, 100-199, 1000-1999, etc.

    **This query will utilize the database index if one exists.**

    :param incomplete: The incomplete string.
    :param lookup_field: The name of the model field to use for lookup.
    :param queryset: The starting queryset to use to determine integer ranges.
    :return: A Q object to use for filtering the queryset.
    :raises ValueError: If the incomplete string is not a valid integer.
    :raises TypeError: If the incomplete string is not a valid integer.
    """
    qry = models.Q()
    neg = incomplete.startswith("-")
    for lower, upper in int_ranges(
        incomplete,
        queryset.aggregate(models.Max(lookup_field))[f"{lookup_field}__max"],
    ):
        qry |= models.Q(
            **{f"{lookup_field}__gt{'' if neg else 'e'}": lower}
        ) & models.Q(**{f"{lookup_field}__lt{'e' if neg else ''}": upper})
    return qry


def float_query(
    incomplete: str, lookup_field: str, queryset: QuerySet, **_
) -> models.Q:
    """
    The default completion query builder for float fields. This method will
    return a Q object that will match any value that starts with the incomplete
    string. For example, if the incomplete string is "1.1", the query will match
    1.1 <= float(incomplete) < 1.2

    **This query will utilize the database index if one exists.**

    :param incomplete: The incomplete string.
    :param lookup_field: The name of the model field to use for lookup.
    :param queryset: The starting queryset to use to determine float ranges.
    :return: A Q object to use for filtering the queryset.
    :raises ValueError: If the incomplete string is not a valid float.
    :raises TypeError: If the incomplete string is not a valid float.
    """
    incomplete = incomplete.rstrip("0").rstrip(".")
    lower = float(incomplete)
    if "." in incomplete:
        upper = lower + float(
            f"0.{'0' * (len(incomplete) - incomplete.index('.') - 2)}1"
        )
    else:
        return int_query(
            incomplete=incomplete, lookup_field=lookup_field, queryset=queryset
        )
    return models.Q(**{f"{lookup_field}__gte": lower}) & models.Q(
        **{f"{lookup_field}__lt": upper}
    )


def text_query(
    incomplete: str, lookup_field: str, case_insensitive: bool = False, **_
) -> models.Q:
    """
    The default completion query builder for text-based fields. This method will
    return a Q object that will match any value that starts with the incomplete
    string. Case sensitivity is determined by the case_insensitive constructor
    parameter.

    **This query will utilize the database index if one exists.**

    :param incomplete: The incomplete string.
    :param lookup_field: The name of the model field to use for lookup.
    :param case_insensitive: If the lookup should be case insensitive or not.
    :return: A Q object to use for filtering the queryset.
    """
    if case_insensitive:
        return models.Q(**{f"{lookup_field}__istartswith": incomplete})
    return models.Q(**{f"{lookup_field}__startswith": incomplete})


def uuid_query(incomplete: str, lookup_field: str, **_) -> t.Tuple[models.Q, int]:
    """
    The default completion query builder for UUID fields. This method will
    return a Q object that will match any value that starts with the incomplete
    string. The incomplete string will be stripped of all non-alphanumeric
    characters and padded with zeros to 32 characters. For example, if the
    incomplete string is "a", the query will match
    a0000000-0000-0000-0000-000000000000 to affffffff-ffff-ffff-ffff-ffffffffffff.

    **This query will utilize the database index if one exists.**

    :param incomplete: The incomplete string.
    :param lookup_field: The name of the model field to use for lookup.
    :return: A 2-tuple where the first element is the Q object to use for filtering the
        queryset and the second is the integer offset into the incomplete string where
        the completion characters should be concatenated.
    :raises ValueError: If the incomplete string is too long or contains invalid
        UUID characters. Anything other than (0-9a-fA-F).
    """
    from uuid import UUID

    # the offset futzing is to allow users to ignore the - in the UUID
    # as a convenience of its implementation any non-alpha numeric character
    # will be ignored, and the completion suggestions and parsing will still work
    uuid = ""
    offset = 0
    for char in incomplete:
        if char.isalnum():
            uuid += char
        else:
            offset -= 1

    if len(incomplete) >= 9:
        offset += 1
    if len(incomplete) >= 14:
        offset += 1
    if len(incomplete) >= 19:
        offset += 1
    if len(incomplete) >= 24:
        offset += 1

    if len(uuid) > 32:
        raise ValueError(f"Too many UUID characters: {incomplete}")
    min_uuid = UUID(uuid + "0" * (32 - len(uuid)))
    max_uuid = UUID(uuid + "f" * (32 - len(uuid)))
    return (
        models.Q(**{f"{lookup_field}__gte": min_uuid})
        & models.Q(**{f"{lookup_field}__lte": max_uuid}),
        offset,
    )


def date_query(incomplete: str, lookup_field: str, **_) -> models.Q:
    """
    Default completion query builder for date fields. This method will return a Q
    object that will match any value that starts with the incomplete date string.
    All dates must be in ISO8601 format (YYYY-MM-DD).

    **This query will utilize the database index if one exists.**

    :param incomplete: The incomplete string.
    :param lookup_field: The name of the model field to use for lookup.
    :return: A Q object to use for filtering the queryset.
    :raises ValueError: If the incomplete string is not a valid partial date.
    :raises AssertionError: If the incomplete string is not a valid partial date.
    """
    lower_bound, upper_bound = get_date_bounds(incomplete)
    return models.Q(**{f"{lookup_field}__gte": lower_bound}) & models.Q(
        **{f"{lookup_field}__lte": upper_bound}
    )


def time_query(incomplete: str, lookup_field: str, **_) -> models.Q:
    """
    Default completion query builder for time fields. This method will return a Q
    object that will match any value that starts with the incomplete time string.
    All times must be in ISO 8601 format (HH:MM:SS.ssssss).

    **This query will utilize the database index if one exists.**

    :param incomplete: The incomplete string.
    :param lookup_field: The name of the model field to use for lookup.
    :return: A Q object to use for filtering the queryset.
    :raises ValueError: If the incomplete string is not a valid partial time.
    :raises AssertionError: If the incomplete string is not a valid partial time.
    """
    lower_bound, upper_bound = get_time_bounds(incomplete)
    return models.Q(**{f"{lookup_field}__gte": lower_bound}) & models.Q(
        **{f"{lookup_field}__lte": upper_bound}
    )


def datetime_query(incomplete: str, lookup_field: str, **_) -> models.Q:
    """
    Default completion query builder for datetime fields. This method will return a
    Q object that will match any value that starts with the incomplete datetime
    string. All dates must be in ISO8601 format (YYYY-MM-DDTHH:MM:SS.ssssss±HH:MM).

    **This query will utilize the database index if one exists.**

    :param incomplete: The incomplete string.
    :param lookup_field: The name of the model field to use for lookup.
    :return: A Q object to use for filtering the queryset.
    :raises ValueError: If the incomplete string is not a valid partial datetime.
    :raises AssertionError: If the incomplete string is not a valid partial
        datetime.
    """
    import re
    from datetime import datetime

    from django.utils.timezone import get_default_timezone, make_aware

    parts = incomplete.split("T")
    lower_bound, upper_bound = get_date_bounds(parts[0])

    def get_tz_part(dt_str: str) -> str:
        return dt_str[dt_str.rindex("+") if "+" in dt_str else dt_str.rindex("-") :]

    time_lower = datetime.min.time()
    time_upper = datetime.max.time()
    tz_part = ""
    if len(parts) > 1:
        time_parts = re.split(r"[+-]", parts[1])
        time_lower, time_upper = get_time_bounds(time_parts[0])
        # we punt on the timezones - if the user supplies a partial timezone
        # different than the default django timezone, its just too complicated to be
        # worth trying to complete, we ensure it aligns as a prefix to the
        # configured default timezone instead
        if len(time_parts) > 1 and parts[1]:
            tz_part = get_tz_part(parts[1])
    lower_bound = datetime.combine(lower_bound, time_lower)
    upper_bound = datetime.combine(upper_bound, time_upper)

    if settings.USE_TZ:
        lower_bound = make_aware(lower_bound, get_default_timezone())
        upper_bound = make_aware(upper_bound, get_default_timezone())
        db_tz_part = get_tz_part(lower_bound.isoformat())
        assert db_tz_part.startswith(tz_part)
    else:
        assert not tz_part
    return models.Q(**{f"{lookup_field}__gte": lower_bound}) & models.Q(
        **{f"{lookup_field}__lte": upper_bound}
    )


def duration_query(
    incomplete: str, lookup_field: str, queryset: QuerySet, **_
) -> models.Q:
    """
    Default completion query builder for duration fields. This method will return a
    Q object that will match any value that is greater than the incomplete duration
    string (or less if negative). Duration strings are formatted in a subset of the
    ISO8601 standard. Only day, hours, minutes and fractional seconds are supported.
    Year, week and month specifiers are not.

    **This query will utilize the database index if one exists.**

    :param incomplete: The incomplete string.
    :param lookup_field: The name of the model field to use for lookup.
    :param queryset: The queryset to use to determine integer ranges.
    :return: A Q object to use for filtering the queryset.
    :raises ValueError: If the incomplete string is not a valid partial duration.
    :raises AssertionError: If the incomplete string is not a valid partial
        duration.
    """
    from django_typer.utils import parse_iso_duration

    duration, ambiguity = parse_iso_duration(incomplete)
    if incomplete.endswith("S") or (duration.microseconds and not ambiguity):
        return models.Q(**{f"{lookup_field}": duration})
    neg = incomplete.startswith("-")

    qry = models.Q()
    horizon = None  # time horizon is exclusive!

    if ambiguity and "T" not in incomplete and "D" not in incomplete:
        # days is unbounded
        # if days == 5, we want to match 5-<6, 50-<60, 500-<600, etc
        max_val = (
            queryset.filter(qry).aggregate(models.Min(lookup_field))[
                f"{lookup_field}__min"
            ]
            if neg
            else queryset.filter(qry).aggregate(models.Max(lookup_field))[
                f"{lookup_field}__max"
            ]
        )
        for lower, upper in int_ranges(
            ambiguity,
            max_val.days,
        ):
            qry |= models.Q(
                **{f"{lookup_field}__gt{'' if neg else 'e'}": timedelta(days=lower)}
            ) & models.Q(
                **{f"{lookup_field}__lt{'e' if neg else ''}": timedelta(days=upper)}
            )
    elif duration.days or "D" in incomplete or "T" in incomplete:
        horizon = timedelta(days=1)

    if "T" in incomplete:
        hours, seconds = divmod(duration.seconds, 3600)
        minutes, seconds = divmod(seconds, 60)

        # if we are here, we may or may not have an ambiguous time component
        # or (exclusively) we may be missing time components
        if ambiguity is None:
            # handle no ambiguity first
            # there's no way to be here and have any ambiguity or non-ambiguous
            # microseconds ex: PT PT1. PT1M PT1H PT1H1M
            if incomplete.endswith("M"):
                horizon = timedelta(minutes=1)
            elif incomplete.endswith("H"):
                horizon = timedelta(hours=1)
            elif not incomplete.endswith("T"):
                horizon = timedelta(seconds=1)
        else:
            # we have a trailing ambiguity
            if "." in incomplete or seconds:
                # microsecond ambiguity
                # 5.000 -> the most this could be is 5.000999
                floor = int(f"{ambiguity:0<6}")
                duration = timedelta(
                    days=abs(duration.days),
                    seconds=abs(duration.seconds),
                    microseconds=floor,
                )
                duration = -duration if neg else duration
                horizon = timedelta(
                    microseconds=int(ambiguity + "9" * (6 - len(ambiguity))) + 1 - floor
                )
            else:
                # ambiguity is at least a seconds ambiguity
                int_amb = int(ambiguity)
                compound_horizon: t.List[
                    t.Tuple[int, int]
                ] = []  # seconds horizons small -> large
                compound_horizon.append(
                    (
                        int_amb,
                        int_amb + 1,
                    )
                )
                compound_horizon.append(
                    (int(f"{ambiguity}0"), min(int(f"{ambiguity}9") + 1, 60))
                )
                if "M" not in incomplete:
                    # ambiguity is minutes or seconds
                    compound_horizon.append(
                        (
                            int_amb * 60,
                            int_amb * 60 + 60,
                        )
                    )
                    if len(ambiguity) == 1:
                        compound_horizon.append(
                            (
                                int(f"{ambiguity}0") * 60,
                                min(int(f"{ambiguity}9") + 1, 60) * 60,
                            )
                        )
                if "H" not in incomplete:
                    # ambiguity is hours or minutes or seconds
                    # bug here T1 could be T1H or T10H-T19H
                    compound_horizon.append(
                        (
                            int_amb * 3600,
                            int_amb * 3600 + 3600,
                        )
                    )
                    if len(ambiguity) == 1:
                        compound_horizon.append(
                            (
                                int(f"{ambiguity}0") * 3600,
                                min(int(f"{ambiguity}9") + 1, 24) * 3600,
                            )
                        )
                c_qry = models.Q()
                for lower, upper in compound_horizon:
                    lwr, upr = (
                        (
                            duration - timedelta(seconds=upper),
                            duration - timedelta(seconds=lower),
                        )
                        if neg
                        else (
                            duration + timedelta(seconds=lower),
                            duration + timedelta(seconds=upper),
                        )
                    )
                    h_qry = models.Q(
                        **{f"{lookup_field}__gt{'' if neg else 'e'}": lwr}
                    ) & models.Q(**{f"{lookup_field}__lt{'e' if neg else ''}": upr})
                    c_qry |= h_qry
                qry &= c_qry

    inclusive = "" if incomplete.endswith("T") and duration.days else "e"
    qry &= (
        models.Q(**{f"{lookup_field}__lt{inclusive}": duration})
        if neg
        else models.Q(**{f"{lookup_field}__gt{inclusive}": duration})
    )

    if horizon:
        qry &= (
            models.Q(**{f"{lookup_field}__gt": duration - horizon})
            if neg
            else models.Q(**{f"{lookup_field}__lt": duration + horizon})
        )
    return qry


def get_date_bounds(incomplete: str) -> t.Tuple[date, date]:
    """
    Turn an incomplete YYYY-MM-DD date string into upper and lower bound date
    objects.

    :param incomplete: The incomplete time string.
    :return: A 2-tuple of (lower, upper) date object boundaries.
    """
    import calendar

    parts = incomplete.split("-")
    year_low = max(int(parts[0] + "0" * (4 - len(parts[0]))), 1)
    year_high = int(parts[0] + "9" * (4 - len(parts[0])))
    month_high = 12
    month_low = 1
    day_low = 1
    day_high = None
    if len(parts) > 1:
        assert len(parts[0]) > 3, "Year must be 4 digits"
        month_high = min(int(parts[1] + "9" * (2 - len(parts[1]))), 12)
        month_low = max(int(parts[1] + "0" * (2 - len(parts[1]))), 1)
        if len(parts) > 2:
            assert len(parts[1]) > 1, "Month must be 2 digits"
            day_low = max(int(parts[2] + "0" * (2 - len(parts[2]))), 1)
            day_high = min(
                int(parts[2] + "9" * (2 - len(parts[2]))),
                calendar.monthrange(year_high, month_high)[1],
            )
    lower_bound = date(year=year_low, month=month_low, day=day_low)
    upper_bound = date(
        year=year_high,
        month=month_high,
        day=day_high or calendar.monthrange(year_high, month_high)[1],
    )
    return lower_bound, upper_bound


def get_time_bounds(incomplete: str) -> t.Tuple[time, time]:
    """
    Turn an incomplete HH::MM::SS.ssssss time string into upper and lower bound time
    objects.

    :param incomplete: The incomplete time string.
    :return: A 2-tuple of (lower, upper) time object boundaries.
    """
    time_parts = incomplete.split(":")
    if time_parts and time_parts[0]:
        hours_low = int(time_parts[0] + "0" * (2 - len(time_parts[0])))
        hours_high = min(int(time_parts[0] + "9" * (2 - len(time_parts[0]))), 23)
        minutes_low = 0
        minutes_high = 59
        seconds_low = 0
        seconds_high = 59
        microseconds_low = 0
        microseconds_high = 999999
        if len(time_parts) > 1:
            assert len(time_parts[0]) > 1  # Hours must be 2 digits
            minutes_low = int(time_parts[1] + "0" * (2 - len(time_parts[1])))
            minutes_high = min(int(time_parts[1] + "9" * (2 - len(time_parts[1]))), 59)
            if len(time_parts) > 2:
                seconds_parts = time_parts[2].split(".")
                int_seconds = seconds_parts[0]
                assert len(time_parts[1]) > 1  # Minutes must be 2 digits
                seconds_low = int(int_seconds + "0" * (2 - len(int_seconds)))
                seconds_high = min(int(int_seconds + "9" * (2 - len(int_seconds))), 59)
                if len(seconds_parts) > 1:
                    microseconds = seconds_parts[1]
                    microseconds_low = int(microseconds + "0" * (6 - len(microseconds)))
                    microseconds_high = int(
                        microseconds + "9" * (6 - len(microseconds))
                    )
        return time(
            hour=hours_low,
            minute=minutes_low,
            second=seconds_low,
            microsecond=microseconds_low,
        ), time(
            hour=hours_high,
            minute=minutes_high,
            second=seconds_high,
            microsecond=microseconds_high,
        )
    return time.min, time.max


class ModelObjectCompleter:
    """
    A completer for generic Django model objects. This completer will work
    for most Django core model field types where completion makes sense.

    This completer currently supports the following field types and their subclasses:

        - :class:`~django.db.models.IntegerField`
            - :class:`~django.db.models.AutoField`
            - :class:`~django.db.models.BigAutoField`
            - :class:`~django.db.models.BigIntegerField`
            - :class:`~django.db.models.SmallIntegerField`
            - :class:`~django.db.models.PositiveIntegerField`
            - :class:`~django.db.models.PositiveSmallIntegerField`
            - :class:`~django.db.models.SmallAutoField`
        - :class:`~django.db.models.CharField`
            - :class:`~django.db.models.SlugField`
            - :class:`~django.db.models.URLField`
            - :class:`~django.db.models.EmailField`
        - :class:`~django.db.models.FileField`
            - :class:`~django.db.models.ImageField`
        - :class:`~django.db.models.FilePathField`
        - :class:`~django.db.models.TextField`
        - :class:`~django.db.models.DateField` **(Must use ISO 8601: YYYY-MM-DD)**
        - :class:`~django.db.models.TimeField` **(Must use ISO 8601: HH:MM:SS.ssssss)**
        - :class:`~django.db.models.DateTimeField` **(Must use ISO 8601: YYYY-MM-DDTHH:MM:SS.ssssss±HH:MM)**
        - :class:`~django.db.models.DurationField` **(Must use ISO 8601: YYYY-MM-DDTHH:MM:SS.ssssss±HH:MM)**
        - :class:`~django.db.models.UUIDField`
        - :class:`~django.db.models.FloatField`
        - :class:`~django.db.models.DecimalField`
        - :class:`~django.db.models.GenericIPAddressField`

    .. note::

        The queries used by this completer will make use of column indexes. Completions
        should be fast even for large data.

    The completer query logic is pluggable, but the defaults cover most use cases. The
    limit field is important. It defaults to 50 meaning if more than 50 potential
    completions are found only the first 50 will be returned and there will be no
    indication to the user that there are more. This is to prevent the shell from
    becoming unresponsive when offering completion for large tables.

    To use this completer, pass an instance of this class to the `shell_complete`
    argument of a typer.Option or typer.Argument:

    .. code-block:: python

        from django_typer.completers import ModelObjectCompleter

        class Command(TyperCommand):

            def handle(
                self,
                model_obj: Annotated[
                    MyModel,
                    typer.Argument(
                        shell_complete=ModelObjectCompleter(MyModel, lookup_field="name"),
                        help=_("The model object to use.")
                    )
                ]
            ):
                ...

    .. note::

        See also :func:`~django_typer.utils.model_parser_completer` for a convenience
        function that returns a configured parser and completer for a model object
        and helps reduce boilerplate.

    :param model_or_qry: The Django model class or a queryset to filter against.
    :param lookup_field: The name of the model field to use for lookup.
    :param help_field: The name of the model field to use for help text or None if
        no help text should be provided.
    :param query: A callable that accepts any named arguments and returns a Q object.
        It will be passed:

        - **incomplete** the incomplete string
        - **lookup_field** the name of the model field to use for lookup
        - **queryset** the base queryset to use for completions
        - **context** the click context
        - **parameter** the click parameter
        - **completer** an instance of this ModelObjectCompleter class

        It is not required to use those arguments, but they are available.

        It must return a Q object to use for filtering the queryset. The default query
        will use the relevant query builder depending on the lookup field class.

        ..note::

            The query builder function may also return a second integer offset value.
            This value will be used to adjust the index into the completion strings when
            we concatenate the incomplete string with the lookup field value - see UUID
            which has to allow for - to be missing
    :param limit: The maximum number of completion items to return. If None, all
        matching items will be returned. When offering completion for large tables
        you'll want to set this to a reasonable limit. Default: 50
    :param case_insensitive: Whether or not to perform case insensitive matching when
        completing text-based fields. Defaults to False.
    :param distinct: Whether or not to filter out duplicate values. Defaults to True.
        This is not the same as calling distinct() on the queryset - which will happen
        regardless - but rather whether or not to filter out values that are already
        given for the parameter on the command line.
    :param order_by: The order_by parameter to prioritize completions in. By default
        the default queryset ordering will be used for the model.
    :param use_choices: Whether or not to use the field choices for completion. If True,
        matches to choice values coerced to strings will be returned. If False, the
        field's default query builder will be used instead.
    """

    QueryBuilder = t.Callable[
        ["ModelObjectCompleter", Context, Parameter, str], models.Q
    ]

    model_cls: t.Type[models.Model]
    queryset: QuerySet
    lookup_field: str
    help_field: t.Optional[str] = None
    query: t.Callable[..., t.Union[models.Q, t.Tuple[models.Q, int]]]
    limit: t.Optional[int] = 50
    case_insensitive: bool = False
    distinct: bool = True
    order_by: t.List[str] = []
    use_choices: bool = True

    _field: models.Field

    @staticmethod
    def to_str(obj: t.Any) -> str:
        """
        Convert the given object into a string suitable for use in completions.
        """
        from datetime import datetime

        from django.utils.timezone import get_default_timezone

        if isinstance(obj, datetime):
            if settings.USE_TZ and get_default_timezone():
                obj = obj.astimezone(get_default_timezone())
            return obj.isoformat()
        elif isinstance(obj, time):
            return obj.isoformat()
        elif isinstance(obj, date):
            return obj.isoformat()
        elif isinstance(obj, timedelta):
            from django_typer.utils import duration_iso_string

            return duration_iso_string(obj)
        return str(obj)

    def __init__(
        self,
        model_or_qry: t.Union[t.Type[models.Model], QuerySet],
        lookup_field: t.Optional[str] = None,
        help_field: t.Optional[str] = help_field,
        query: t.Optional[QueryBuilder] = None,
        limit: t.Optional[int] = limit,
        case_insensitive: bool = case_insensitive,
        distinct: bool = distinct,
        order_by: t.Optional[t.Union[str, t.Sequence[str]]] = order_by,
        use_choices: bool = use_choices,
    ):
        import inspect

        if inspect.isclass(model_or_qry) and issubclass(model_or_qry, models.Model):
            self.model_cls = model_or_qry
            self.queryset = model_or_qry.objects.all()
        elif isinstance(model_or_qry, QuerySet):
            self.model_cls = model_or_qry.model
            self.queryset = model_or_qry
        else:
            raise ValueError(
                "ModelObjectCompleter requires a Django model class or queryset."
            )
        self.lookup_field = str(
            lookup_field or getattr(self.model_cls._meta.pk, "name", "id")
        )
        self.help_field = help_field
        self.limit = limit
        self.case_insensitive = case_insensitive
        self.distinct = distinct
        if order_by:
            self.order_by = [order_by] if isinstance(order_by, str) else list(order_by)
        self.use_choices = use_choices

        self._field = self.model_cls._meta.get_field(self.lookup_field)
        if query:
            self.query = query
        else:
            if isinstance(self._field, models.IntegerField):
                self.query = int_query
            elif isinstance(
                self._field,
                (
                    models.CharField,
                    models.TextField,
                    models.GenericIPAddressField,
                    models.FileField,
                    models.FilePathField,
                ),
            ):
                self.query = partial(
                    text_query,
                    case_insensitive=self.case_insensitive,
                )
            elif isinstance(self._field, models.UUIDField):
                self.query = uuid_query
            elif isinstance(self._field, (models.FloatField, models.DecimalField)):
                self.query = float_query
            elif isinstance(self._field, models.DateTimeField):
                self.query = datetime_query
            elif isinstance(self._field, models.DateField):
                self.query = date_query
            elif isinstance(self._field, models.TimeField):
                self.query = time_query
            elif isinstance(self._field, models.DurationField):
                self.query = duration_query
            else:
                raise ValueError(
                    f"Unsupported lookup field class: {self._field.__class__.__name__}"
                )

    def __call__(
        self, context: Context, parameter: Parameter, incomplete: str
    ) -> t.List[CompletionItem]:
        """
        The completer method. This method will return a list of CompletionItem
        objects. If the help_field constructor parameter is not None, the help
        text will be set on the CompletionItem objects. The configured query
        method will be used to filter the queryset. distinct() will also be
        applied and if the distinct constructor parameter is True, values already
        present for the parameter on the command line will be filtered out.

        :param context: The click context.
        :param parameter: The click parameter.
        :param incomplete: The incomplete string.
        :return: A list of CompletionItem objects.
        """

        completion_qry = models.Q(**{self.lookup_field + "__isnull": False})

        offset = 0
        if incomplete:
            try:
                result = self.query(
                    incomplete=incomplete,
                    lookup_field=self.lookup_field,
                    queryset=self.queryset,
                    context=context,
                    parameter=parameter,
                    completer=self,
                )
                if isinstance(result, tuple):
                    completion_qry &= result[0]
                    offset = result[1] if len(result) > 1 else 0
                else:
                    completion_qry &= result
            except (ValueError, TypeError, AssertionError):
                return []

        columns = [self.lookup_field]
        if self.help_field:
            columns.append(self.help_field)

        excluded: t.List[models.Model] = []
        if (
            self.distinct
            and parameter.name
            and context.get_parameter_source(parameter.name)
            is not ParameterSource.DEFAULT
        ):
            excluded = context.params.get(parameter.name, []) or []

        qryset = self.queryset.filter(completion_qry).exclude(
            pk__in=[ex.pk for ex in excluded]
        )
        if self.order_by:
            qryset = qryset.order_by(*self.order_by)

        completions = []
        for values in qryset.distinct().values_list(*columns)[0 : self.limit]:
            str_value = self.to_str(values[0])
            if str_value:
                completions.append(
                    CompletionItem(
                        # use the incomplete string prefix incase this was a case
                        # insensitive match
                        value=incomplete + str_value[len(incomplete) + offset :],
                        help=values[1] if len(values) > 1 else None,
                    )
                )
        return completions



"""
Completers that complete items involving Django's settings.
"""

from django.conf import settings

from . import these_strings

setting = these_strings(
    lambda: (setting for setting in dir(settings) if setting.isupper()),
    allow_duplicates=False,
)
"""
Completes Django :ref:`ref/settings:settings` names. Duplicates are not allowed.
"""

languages = these_strings(
    lambda: getattr(settings, "LANGUAGES", []) or [],
    allow_duplicates=False,
)
"""
Completes Django language codes using :setting:`LANGUAGES`. Duplicates are not allowed.
"""



import os
import re
import pytest


def _marker_in_markexpr(markexpr: str, marker: str) -> bool:
    # cheap/ok heuristic: whole-word match in the expression
    return bool(re.search(rf"(?<!\w){re.escape(marker)}(?!\w)", markexpr or ""))


@pytest.hookimpl(tryfirst=True)
def pytest_load_initial_conftests(early_config, parser, args):
    """toggle rich on/off based on test markers"""
    markexpr = ""
    for i, a in enumerate(args):
        if a == "-m" and i + 1 < len(args):
            markexpr = args[i + 1]
            break
        if a.startswith("-m") and len(a) > 2:
            markexpr = a[2:]
            break

    # Toggle rich early based on explicit "no_rich" / "rich" markers in the -m expression
    if not _marker_in_markexpr(markexpr, "no_rich"):
        if not _marker_in_markexpr(markexpr, "rich"):
            return
        else:
            use_rich = True
    else:
        use_rich = False

    os.environ["TYPER_USE_RICH"] = "1" if use_rich else "0"
    from tests import utils

    utils.rich_installed = use_rich



import typing as t

from django.utils.translation import gettext_lazy as _
from typer import Argument, Option

from django_typer.management import Typer
from django_typer.completers.model import ModelObjectCompleter
from django_typer.parsers.model import ModelObjectParser
from tests.apps.examples.polls.models import Question as Poll


app = Typer()


@app.command()
def handle(
    self,
    polls: t.Annotated[
        t.List[Poll],
        Argument(
            parser=ModelObjectParser(Poll),
            shell_complete=ModelObjectCompleter(
                Poll, help_field="question_text"
            ),
            help=_("The database IDs of the poll(s) to close."),
        ),
    ],
    delete: t.Annotated[
        bool,
        Option(
            "--delete",  # we can also get rid of that unnecessary --no-delete flag
            help=_("Delete poll instead of closing it."),
        ),
    ] = False,
):
    """
    Closes the specified poll for voting.


    As mentioned in the last section, helps can also
    be set in the docstring
    """
    for poll in polls:
        poll.opened = False
        poll.save()
        self.stdout.write(
            self.style.SUCCESS(f'Successfully closed poll "{poll.id}"')
        )
        if delete:
            poll.delete()



from django.core.management import CommandError, call_command
from django.test import TestCase, override_settings

from django_typer.management import get_command
from tests.utils import run_command


class ModifyOnInitializeTests(TestCase):
    """
    Tests that you can build a command programmatically when it is instantiated.
    """

    def test_mod_on_init_run(self):
        self.assertEqual(
            run_command(
                "mod_on_init",
                "--settings",
                "tests.settings.mod_init",
                "routine2",
                "--package",
                "--deploy",
            )[0].strip(),
            "('routine2', ['deploy', 'package'])",
        )
        self.assertEqual(
            run_command(
                "mod_on_init",
                "--settings",
                "tests.settings.mod_init",
                "routine3",
            )[0].strip(),
            "('routine3', [])",
        )
        self.assertEqual(
            run_command(
                "mod_on_init",
                "--settings",
                "tests.settings.mod_init",
                "routine3",
                "--deploy",
            )[1].strip(),
            "No such option: --deploy",
        )

    @override_settings(
        ROUTINES={
            "routine1": ["initial", "destroy"],
            "routine2": ["deploy", "package", "upload"],
            "routine3": [],
        }
    )
    def test_mod_on_init_call(self):
        self.assertEqual(
            call_command("mod_on_init", "routine2", package=True, deploy=True),
            ("routine2", ["deploy", "package"]),
        )
        self.assertEqual(call_command("mod_on_init", "routine3"), ("routine3", []))
        with self.assertRaises(CommandError):
            call_command("mod_on_init", "routine3", "--deploy")

    @override_settings(
        ROUTINES={
            "routine1": ["initial", "destroy"],
            "routine2": ["deploy", "package", "upload"],
            "routine3": [],
        }
    )
    def test_mod_on_init_direct(self):
        mod_on_init = get_command("mod_on_init")
        # make sure __init__ is run more than once
        mod_on_init = get_command("mod_on_init")
        mod_on_init = get_command("mod_on_init")
        self.assertEqual(len(mod_on_init.typer_app.registered_commands), 3)
        self.assertEqual(
            mod_on_init.routine2(package=True, upload=True),
            ("routine2", ["package", "upload"]),
        )



import os
import sys
import typing as t
from functools import partial
from pathlib import Path

from click import Context, Parameter
from click.shell_completion import CompletionItem
from django.conf import settings


def _settings_path(name: str) -> t.Optional[Path]:
    s_pth = getattr(settings, name, None)
    if s_pth:
        return Path(s_pth)
    return None


def import_paths(
    ctx: Context,
    param: Parameter,
    incomplete: str,
    root: t.Union[t.Callable[[], t.Optional[Path]], t.Optional[Path]] = None,
) -> t.List[CompletionItem]:
    """
    A completer that completes a python dot import path string based on sys.path.

    :param ctx: The click context.
    :param param: The click parameter.
    :param incomplete: The incomplete string.
    :param root: The root path to search for modules.
    :return: A list of available matching import paths
    """
    import pkgutil

    rt = root() if callable(root) else root
    incomplete = incomplete.strip()
    completions = []
    packages = [pkg for pkg in incomplete.split(".") if pkg]
    pkg_complete = not incomplete or incomplete.endswith(".")
    module_import = ".".join(packages) if pkg_complete else ".".join(packages[:-1])
    module_path = Path(module_import.replace(".", "/"))
    search_paths = []

    if rt and (rt / module_path).exists():
        search_paths.append(str(rt / module_path))
    else:
        for pth in sys.path:
            if (Path(pth) / module_path).exists():
                search_paths.append(str(Path(pth) / module_path))

    prefix = "" if pkg_complete else packages[-1]
    for module in pkgutil.iter_modules(path=search_paths):
        if module.name.startswith(prefix):
            completions.append(
                CompletionItem(
                    f"{module_import}{'.' if module_import else ''}{module.name}",
                    type="plain",
                )
            )
    if len(completions) == 1 and not completions[0].value.endswith("."):
        return import_paths(ctx, param, f"{completions[0].value}.") or completions
    return completions


def paths(
    ctx: Context,
    param: Parameter,
    incomplete: str,
    dir_only: t.Optional[bool] = None,
    root: t.Union[t.Callable[[], t.Optional[Path]], t.Optional[Path]] = None,
) -> t.List[CompletionItem]:
    """
    A completer that completes a path. Relative incomplete paths are interpreted
    relative to the current working directory.

    :param ctx: The click context.
    :param param: The click parameter.
    :param incomplete: The incomplete string.
    :param dir_only: Restrict completions to paths to directories only, otherwise
        complete directories or files.
    :param root: Restrict completions to this root path.
    :return: A list of available matching directories
    """
    rt = root() if callable(root) else root

    def exists(pth: Path) -> bool:
        if dir_only:
            return pth.is_dir()
        return pth.exists() or pth.is_symlink()

    separator = os.sep
    if "/" in incomplete:
        if "\\" not in incomplete:
            separator = "/"
    elif "\\" in incomplete:
        separator = "\\"

    completions = []
    if rt:
        incomplete_path = rt / Path(
            incomplete.replace(separator, os.path.sep).lstrip(os.path.sep)
        )
    else:
        incomplete_path = Path(incomplete.replace(separator, os.path.sep))
    partial_dir = ""
    if not exists(incomplete_path) and not incomplete.endswith(separator):
        partial_dir = incomplete_path.name
        incomplete_path = incomplete_path.parent
    elif incomplete_path.is_file() and not dir_only:
        return [CompletionItem(incomplete, type="file")]
    if incomplete_path.is_dir():
        for child in os.listdir(incomplete_path):
            if not exists(incomplete_path / child):
                continue
            if child.startswith(partial_dir):
                to_complete = incomplete[0 : (-len(partial_dir) or None)]
                sep = (
                    ""
                    if not to_complete or to_complete.endswith(separator)
                    else separator
                )
                ctype = (
                    "plain"
                    if rt
                    else "dir"
                    if (incomplete_path / child).is_dir()
                    else "file"
                )
                completions.append(
                    CompletionItem(
                        f"{to_complete}{sep}{child}",
                        type=ctype,
                    )
                )
    if (
        len(completions) == 1
        and Path(completions[0].value).is_dir()
        and [
            child
            for child in os.listdir(completions[0].value)
            if exists(Path(completions[0].value) / child)
        ]
    ):
        # recurse because we can go futher
        return paths(ctx, param, completions[0].value, dir_only=dir_only)
    return completions


directories = partial(paths, dir_only=True)
"""
A completer that completes a directory path (but not files). Relative incomplete paths
are interpreted relative to the current working directory.

:param ctx: The click context.
:param param: The click parameter.
:param incomplete: The incomplete string.
:return: A list of available matching directories
"""


static_paths = partial(paths, root=partial(_settings_path, name="STATIC_ROOT"))
"""
Complete static file paths.

:param ctx: The click context.
:param param: The click parameter.
:param incomplete: The incomplete string.
:param dir_only: Restrict completions to paths to directories only, otherwise
    complete directories or files.
:return: A list of available matching directories
"""

media_paths = partial(paths, root=partial(_settings_path, name="MEDIA_ROOT"))
"""
Complete media file paths.

:param ctx: The click context.
:param param: The click parameter.
:param incomplete: The incomplete string.
:param dir_only: Restrict completions to paths to directories only, otherwise
    complete directories or files.
:return: A list of available matching directories
"""



from django_typer.management import TyperCommand, command


# here we pass chain=True to typer telling it to allow invocation of
# multiple subcommands
class Command(TyperCommand, chain=True):
    @command()
    def cmd1(self):
        self.stdout.write("cmd1")

    @command()
    def cmd2(self):
        self.stdout.write("cmd2")



from django_typer.management import Typer


def to_csv(results, **_):
    # result_callback is passed the CLI parameters on the current context
    # if we are uninterested in them, we can use the **_ syntax to ignore them
    return ", ".join(results)


app = Typer(result_callback=to_csv, chain=True)


@app.command()
def cmd1():
    return "result1"


@app.command()
def cmd2():
    return "result2"



from django_typer.management import TyperCommand
import json
import sys


class Command(TyperCommand):
    requires_system_checks = []
    requires_migrations_checks = False

    def handle(self, test_arg: int, print: bool = False):
        if print:
            return json.dumps(
                {
                    "typer": test_arg,
                    "print": print,
                    "modules": list(sys.modules.keys()),
                }
            )



from django_typer.management import Typer, DTGroup
from click import Context
import typing as t


class AlphabetizeCommands(DTGroup):
    def list_commands(self, ctx: Context) -> t.List[str]:
        return list(sorted(self.commands.keys()))


app = Typer(cls=AlphabetizeCommands)

d_app = Typer(cls=AlphabetizeCommands)
app.add_typer(d_app, name="d")


@app.command()
def b():
    print("b")


@app.command()
def a():
    print("a")


@d_app.callback()
def d():
    print("d")


@d_app.command()
def f():
    print("f")


@d_app.command()
def e():
    print("e")


@app.command()
def c():
    print("c")



from django_typer.management import TyperCommand, command, initialize


class Command(TyperCommand):
    @initialize(invoke_without_command=True)
    def init(self):
        assert self.__class__ is Command
        return "test_app::init()"

    @command()
    def cmd1(self):
        """
        Command1
        """
        assert self.__class__ is Command
        return "test_app::cmd1()"

    @TyperCommand.command()  # overly verbose but tolerated
    def cmd2(self):
        """
        Command2
        """
        assert self.__class__ is Command
        return "test_app::cmd2()"



from io import StringIO

import pytest
from django.core.management import call_command
from django.test import TestCase, override_settings
from contextlib import redirect_stdout

from django_typer.management import TyperCommand, get_command
from tests.utils import rich_installed, run_command, similarity


@override_settings(INSTALLED_APPS=["tests.apps.howto"])
class TestDefaultCmdHowto(TestCase):
    cmd = "default_cmd"

    def test_howto_default_cmd(self):
        from tests.apps.howto.management.commands.default_cmd import (
            Command,
        )

        command = get_command(self.cmd, Command)
        self.assertEqual(command(), "handle")
        self.assertEqual(command.subcommand2(), "subcommand2")
        self.assertEqual(command.subcommand3(), "subcommand3")

        with self.assertRaises(Exception):
            command.handle()

        self.assertEqual(command.subcommand1(), "handle")


class TestDefaultCmdTyperHowto(TestDefaultCmdHowto):
    cmd = "default_cmd_typer"


@pytest.mark.rich
@pytest.mark.skipif(not rich_installed, reason="This test requires rich help output")
@override_settings(INSTALLED_APPS=["tests.apps.howto"])
class TestGroupsHowto(TestCase):
    cmd = "groups"

    root_help = """
 Usage: ./howto.py groups [OPTIONS] COMMAND [ARGS]...                           
                                                                                
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Django ─────────────────────────────────────────────────────────────────────╮
│ --version                  Show program's version number and exit.           │
│ --settings           TEXT  The Python path to a settings module, e.g.        │
│                            "myproject.settings.main". If this isn't          │
│                            provided, the DJANGO_SETTINGS_MODULE environment  │
│                            variable will be used.                            │
│ --pythonpath         PATH  A directory to add to the Python path, e.g.       │
│                            "/home/djangoprojects/myproject".                 │
│ --traceback                Raise on CommandError exceptions                  │
│ --show-locals              Print local variables in tracebacks.              │
│ --no-color                 Don't colorize the command output.                │
│ --force-color              Force colorization of the command output.         │
│ --skip-checks              Skip system checks.                               │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Commands ───────────────────────────────────────────────────────────────────╮
│ group1                                                                       │
│ group2                                                                       │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

    group1_help = """
 Usage: ./howto.py groups group1 [OPTIONS] COMMAND [ARGS]...                    
                                                                                
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --common-option    --no-common-option      [default: no-common-option]       │
│ --help                                     Show this message and exit.       │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Commands ───────────────────────────────────────────────────────────────────╮
│ grp1-subcommand1                                                             │
│ grp1-subcommand2                                                             │
│ subgroup1                                                                    │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

    subgroup1_help = """
 Usage: ./howto.py groups group1 subgroup1 [OPTIONS] COMMAND [ARGS]...          
                                                                                
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Commands ───────────────────────────────────────────────────────────────────╮
│ subgrp-command                                                               │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

    def test_howto_groups(self):
        stdout = StringIO()

        groups = get_command(self.cmd, TyperCommand, stdout=stdout, no_color=True)

        groups.print_help("./howto.py", "groups")
        self.assertGreaterEqual(
            similarity(stdout.getvalue().strip(), self.root_help.strip()), 0.99
        )
        stdout.truncate(0)
        stdout.seek(0)

        groups.print_help("./howto.py", "groups", "group1")
        self.assertGreaterEqual(
            similarity(stdout.getvalue().strip(), self.group1_help.strip()), 0.99
        )

        stdout.truncate(0)
        stdout.seek(0)

        groups.print_help("./howto.py", "groups", "group1", "subgroup1")
        self.assertGreaterEqual(
            similarity(stdout.getvalue().strip(), self.subgroup1_help.strip()), 0.99
        )

        groups.group1()
        groups.group1.subgroup1()
        groups.group1.subgroup1.subgrp_command()
        groups.group1.grp1_subcommand1()
        groups.group2()

        # also since names are unique this works:
        groups.grp1_subcommand1()
        groups.subgrp_command()


class TestGroupsTyperHowto(TestGroupsHowto):
    cmd = "groups_typer"


@override_settings(INSTALLED_APPS=["tests.apps.howto"])
class TestInitializerHowto(TestCase):
    cmd = "initializer"

    def test_howto_initializer(self):
        from tests.apps.howto.management.commands.initializer import (
            Command,
        )

        command = get_command("initializer", Command)
        command.init(common_option=True)
        self.assertTrue(command.subcommand1())
        command.init(False)
        self.assertFalse(command.subcommand2())


class TestInitializerTyperHowto(TestInitializerHowto):
    cmd = "initializer_typer"


@pytest.mark.rich
@pytest.mark.skipif(not rich_installed, reason="This test requires rich help output")
@override_settings(INSTALLED_APPS=["tests.apps.howto"])
class TestDefaultOptionsHowto(TestCase):
    cmd = "default_options"

    cmd_help = """
 Usage: ./howto.py default_options [OPTIONS]                                    
                                                                                
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Django ─────────────────────────────────────────────────────────────────────╮
│ --verbosity          INTEGER RANGE [0<=x<=3]  Verbosity level; 0=minimal     │
│                                               output, 1=normal output,       │
│                                               2=verbose output, 3=very       │
│                                               verbose output                 │
│                                               [default: 1]                   │
│ --version                                     Show program's version number  │
│                                               and exit.                      │
│ --pythonpath         PATH                     A directory to add to the      │
│                                               Python path, e.g.              │
│                                               "/home/djangoprojects/myproje… │
│ --traceback                                   Raise on CommandError          │
│                                               exceptions                     │
│ --show-locals                                 Print local variables in       │
│                                               tracebacks.                    │
│ --no-color                                    Don't colorize the command     │
│                                               output.                        │
│ --force-color                                 Force colorization of the      │
│                                               command output.                │
│ --skip-checks                                 Skip system checks.            │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

    def test_howto_default_options(self):
        stdout = StringIO()

        groups = get_command(self.cmd, TyperCommand, stdout=stdout, no_color=True)

        groups.print_help("./howto.py", "default_options")
        self.assertGreaterEqual(
            similarity(stdout.getvalue().strip(), self.cmd_help.strip()), 0.99
        )


class TestDefaultOptionsTyperHowto(TestDefaultOptionsHowto):
    cmd = "default_options_typer"


@override_settings(INSTALLED_APPS=["tests.apps.howto"])
class TestConfigureHowto(TestCase):
    cmd = "configure"

    def test_howto_configure(self):
        stdout = StringIO()
        call_command(
            get_command(self.cmd, stdout=stdout, no_color=True), "cmd1", "cmd2"
        )
        self.assertEqual(stdout.getvalue().strip().splitlines()[0:2], ["cmd1", "cmd2"])


class TestConfigureTyperHowto(TestConfigureHowto):
    cmd = "configure_typer"


@override_settings(INSTALLED_APPS=["tests.apps.howto"])
class TestInheritHowto(TestCase):
    cmd1 = "upstream"
    cmd2 = "downstream"

    def test_howto_inherit(self):
        from tests.apps.howto.management.commands import (
            downstream,
            upstream,
        )

        upstream = get_command(
            self.cmd1, upstream.Command, stdout=StringIO(), no_color=True
        )
        downstream = get_command(
            self.cmd2, downstream.Command, stdout=StringIO(), no_color=True
        )
        self.assertEqual(upstream.init(), "upstream:init")
        self.assertEqual(downstream.init(), "downstream:init")

        self.assertEqual(upstream.sub1(), "upstream:sub1")
        self.assertEqual(downstream.sub1(), "downstream:sub1")

        self.assertEqual(upstream.sub2(), "upstream:sub2")
        self.assertEqual(downstream.sub2(), "upstream:sub2")

        self.assertEqual(upstream.grp1(), "upstream:grp1")
        self.assertEqual(downstream.grp1(), "upstream:grp1")

        self.assertEqual(upstream.grp1_cmd1(), "upstream:grp1_cmd1")
        self.assertEqual(downstream.grp1_cmd1(), "upstream:grp1_cmd1")
        self.assertEqual(upstream.grp1.grp1_cmd1(), "upstream:grp1_cmd1")
        self.assertEqual(downstream.grp1.grp1_cmd1(), "upstream:grp1_cmd1")

        self.assertEqual(downstream.sub3(), "downstream:sub3")
        self.assertEqual(downstream.grp1_cmd2(), "downstream:grp1_cmd2")
        self.assertEqual(downstream.grp1.grp1_cmd2(), "downstream:grp1_cmd2")

        with self.assertRaises(AttributeError):
            upstream.sub3()

        with self.assertRaises(AttributeError):
            upstream.grp1_cmd2()

        with self.assertRaises(AttributeError):
            upstream.grp1.grp1_cmd2()


class TestInheritTyperHowto(TestInheritHowto):
    cmd1 = "upstream_typer"
    cmd2 = "downstream_typer"


@override_settings(INSTALLED_APPS=["tests.apps.howto"])
class TestPluginHowto(TestCase):
    cmd = "upstream2"

    def test_howto_plugin(self):
        from tests.apps.howto.management.commands import upstream2

        upstream = get_command(
            self.cmd, upstream2.Command, stdout=StringIO(), no_color=True
        )
        self.assertEqual(upstream.init(), "plugin:init")
        self.assertEqual(upstream.sub1(), "plugin:sub1")
        self.assertEqual(upstream.sub2(), "upstream:sub2")
        self.assertEqual(upstream.grp1(), "upstream:grp1")
        self.assertEqual(upstream.grp1_cmd1(), "upstream:grp1_cmd1")
        self.assertEqual(upstream.sub3(), "plugin:sub3")
        self.assertEqual(upstream.grp1_cmd2(), "plugin:grp1_cmd2")
        self.assertEqual(upstream.grp1.grp1_cmd2(), "plugin:grp1_cmd2")


class TestPluginTyperHowto(TestPluginHowto):
    cmd = "upstream2_typer"


@override_settings(INSTALLED_APPS=["tests.apps.howto"])
class TestPrintingHowto(TestCase):
    cmd = "printing"

    def test_howto_printing(self):
        from tests.apps.howto.management.commands.printing import (
            Command,
        )

        stdout = StringIO()
        get_command(self.cmd, Command, stdout=stdout, no_color=True)()
        self.assertEqual(
            stdout.getvalue().strip().splitlines(),
            ["echo does not support styling", "but secho does!"],
        )

        stdout = StringIO()
        get_command(self.cmd, Command, stdout=stdout, force_color=True)()
        self.assertEqual(
            stdout.getvalue().strip().splitlines(),
            ["echo does not support styling", "\x1b[32mbut secho does!\x1b[0m"],
        )


class TestPrintingTyperHowto(TestPrintingHowto):
    cmd = "printing_typer"


@override_settings(INSTALLED_APPS=["tests.apps.howto"])
class TestOrderHowTo(TestCase):
    cmd = "order"

    from tests.apps.howto.management.commands.order import AlphabetizeCommands

    grp_cls = AlphabetizeCommands

    @pytest.mark.rich
    @pytest.mark.no_rich
    def test_howto_order(self):
        from tests.apps.howto.management.commands.order import Command as OrderCommand

        buffer = StringIO()
        order_cmd = get_command(self.cmd, OrderCommand, stdout=buffer, no_color=True)

        self.assertTrue(issubclass(order_cmd.typer_app.info.cls, self.grp_cls))
        self.assertTrue(issubclass(order_cmd.d.info.cls, self.grp_cls))

        order_cmd.print_help("./manage.py", self.cmd)
        hlp = buffer.getvalue()

        if rich_installed:
            self.assertTrue(
                hlp.index("│ a")
                < hlp.index("│ b")
                < hlp.index("│ c")
                < hlp.index("│ d")
            )
        else:
            cmd_idx = hlp.index("Commands")
            self.assertTrue(
                hlp.index(" a", cmd_idx)
                < hlp.index(" b", cmd_idx)
                < hlp.index(" c", cmd_idx)
                < hlp.index(" d", cmd_idx)
            )

        buffer.seek(0)
        buffer.truncate()

        order_cmd.print_help("./manage.py", self.cmd, "d")
        hlp = buffer.getvalue()

        if rich_installed:
            self.assertTrue(hlp.index("│ e") < hlp.index("│ f"))
        else:
            cmd_idx = hlp.index("Commands")
            self.assertTrue(hlp.index(" e", cmd_idx) < hlp.index(" f", cmd_idx))


class TestPrintingTyperHowto(TestOrderHowTo):
    cmd = "order_typer"

    from tests.apps.howto.management.commands.order_typer import AlphabetizeCommands

    grp_cls = AlphabetizeCommands


@override_settings(INSTALLED_APPS=["tests.apps.howto"])
class TestFinalizerHowto(TestCase):
    command = "finalize"

    def test_howto_finalizer_run(self):
        self.assertEqual(
            run_command(
                self.command,
                "--settings",
                "tests.settings.howto",
                "cmd1",
                "cmd1",
                "cmd2",
            )[0].strip(),
            "result1, result1, result2",
        )

    def test_howto_finalizer_call(self):
        self.assertEqual(
            call_command(self.command, "cmd1", "cmd1", "cmd2"),
            "result1, result1, result2",
        )

    def test_howto_finalizer_obj(self):
        from tests.apps.howto.management.commands.finalize import Command

        command = get_command(self.command, Command)
        self.assertEqual(
            command.to_csv([command.cmd1(), command.cmd1(), command.cmd2()]),
            "result1, result1, result2",
        )


class TestFinalizerTyperHowto(TestFinalizerHowto):
    command = "finalize_typer"

    def test_howto_finalizer_obj(self):
        from tests.apps.howto.management.commands.finalize_typer import Command, to_csv

        command = get_command(self.command, Command)
        self.assertEqual(
            to_csv([command.cmd1(), command.cmd1(), command.cmd2()]),
            "result1, result1, result2",
        )


class TestFinalizerTyperExtHowto(TestFinalizerHowto):
    command = "finalize_typer_ext"


@override_settings(INSTALLED_APPS=["tests.apps.howto"])
class TestFinalizerGroupHowto(TestCase):
    command = "finalize_group"

    def test_howto_finalizer_run(self):
        self.assertEqual(
            run_command(
                self.command,
                "--settings",
                "tests.settings.howto",
                "cmd1",
                "cmd1",
                "cmd2",
                "grp",
                "cmd4",
                "cmd3",
            )[0].strip(),
            "result1, result1, result2, RESULT4, RESULT3",
        )

    def test_howto_finalizer_call(self):
        self.assertEqual(
            call_command(self.command, "cmd1", "cmd1", "cmd2", "grp", "cmd4", "cmd3"),
            "result1, result1, result2, RESULT4, RESULT3",
        )

    def test_howto_finalizer_obj(self):
        from tests.apps.howto.management.commands.finalize_group import Command

        command = get_command(self.command, Command)
        self.assertEqual(
            command.to_csv(
                [
                    command.cmd1(),
                    command.cmd1(),
                    command.cmd2(),
                    command.grp.to_upper_csv([command.grp.cmd4(), command.grp.cmd3()]),
                ]
            ),
            "result1, result1, result2, RESULT4, RESULT3",
        )


class TestFinalizerGroupTyperHowto(TestFinalizerGroupHowto):
    command = "finalize_group_typer"

    def test_howto_finalizer_obj(self):
        from tests.apps.howto.management.commands.finalize_group_typer import (
            Command,
            to_csv,
            to_upper_csv,
        )

        command = get_command(self.command, Command)
        self.assertEqual(
            to_csv(
                [
                    command.cmd1(),
                    command.cmd1(),
                    command.cmd2(),
                    to_upper_csv([command.cmd4(), command.cmd3()]),
                ]
            ),
            "result1, result1, result2, RESULT4, RESULT3",
        )


class TestFinalizerGroupTyperExtHowto(TestFinalizerGroupHowto):
    command = "finalize_group_typer_ext"


@override_settings(INSTALLED_APPS=["tests.apps.howto"])
class TestPrintResultHowTo(TestCase):
    command = "print_result"

    def test_howto_print_result_run(self):
        self.assertEqual(
            run_command(self.command, "--settings", "tests.settings.howto")[0].strip(),
            "",
        )

    def test_howto_print_result_call(self):
        output = StringIO()
        with redirect_stdout(output):
            self.assertEqual(call_command(self.command), "This will not be printed")
        self.assertEqual(output.getvalue().strip(), "")

    def test_howto_print_result_obj(self):
        command = get_command(self.command)
        self.assertEqual(
            command.handle(),
            "This will not be printed",
        )


class TestPrintResultTyperHowTo(TestPrintResultHowTo):
    command = "print_result_typer"



from django.utils import timezone

from django_typer.management import TyperCommand
from tests.apps.examples.polls.models import Question as Poll


class Command(TyperCommand):
    help = "Create some test polls"

    def handle(self):
        self.q1 = Poll.objects.create(
            question_text="What is your favorite ice cream?",
            pub_date=timezone.now(),
        )
        self.q2 = Poll.objects.create(
            question_text="Messi or Ronaldo?",
            pub_date=timezone.now(),
        )
        self.q3 = Poll.objects.create(
            question_text="What about quests?!",
            pub_date=timezone.now(),
        )



import typing as t

from django.utils.translation import gettext_lazy as _
from typer import Argument

from django_typer.management import Typer

app = Typer(help="A command that defines subcommands.")


@app.command()
def create(
    name: t.Annotated[
        str, Argument(help=_("The name of the object to create."))
    ],
):
    """
    Create an object.
    """


@app.command()
def delete(
    id: t.Annotated[int, Argument(help=_("The id of the object to delete."))],
):
    """
    Delete an object.
    """



from django_typer.shells import register_completion_class
from django_typer.shells.bash import BashComplete
from django_typer.shells.fish import FishComplete
from django_typer.shells.powershell import (
    PowerShellComplete,
    PwshComplete,
)
from django_typer.shells.zsh import ZshComplete

register_completion_class(ZshComplete)
register_completion_class(BashComplete)
register_completion_class(PowerShellComplete)
register_completion_class(PwshComplete)
register_completion_class(FishComplete)



import json


from .help_precedence8 import Command as BaseCommand


# if you really want to use a class docstring to override a base class help at
# higher precedence you can unset the base class help attribute by setting the
# higher precedence values to empty strings
class Command(BaseCommand, help=None):
    """
    Now class docstring is used!
    """

    help = None

    def handle(self, arg1: str, arg2: str, arg3: float = 0.5, arg4: int = 1):
        assert self.__class__ is Command
        opts = {
            "arg1": arg1,
            "arg2": arg2,
            "arg3": arg3,
            "arg4": arg4,
            "class": str(self.__class__),
        }
        return json.dumps(opts)



from django_typer.management import Typer, TyperCommand

Command: TyperCommand

app = Typer()

# safe to use Command after first Typer() call
Command.suppressed_base_arguments = {
    "--skip-checks",
    "traceback",
    "force_color",
    "--show-locals",
}


@app.command()
def main(name: str):
    return {"name": name}



from django_typer.management import TyperCommand
from django_typer.types import Verbosity


class Command(TyperCommand):
    suppressed_base_arguments = ["--settings"]  # remove the --settings option

    def handle(self, verbosity: Verbosity = 1): ...



"""
Generated by 'django-admin startproject' using Django 3.2.22.

For more information on this file, see
https://docs.djangoproject.com/en/3.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.2/ref/settings/
"""

import os
from pathlib import Path
import django

try:
    import django_stubs_ext

    django_stubs_ext.monkeypatch()
except ImportError:
    pass

# Set default terminal width for help outputs - necessary for
# testing help output in CI environments.
os.environ["TERMINAL_WIDTH"] = "80"

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-$druy$m$nio-bkagw_%=@(1w)q0=k^mk_5sfk3zi9#4v!%mh*u"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    "tests.apps.completion",
    "tests.apps.test_app",
    "tests.apps.util",
    "django_typer",
    "tests.apps.examples.polls.apps.PollsConfig",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
]

if django.VERSION[0] >= 5:
    INSTALLED_APPS = ["tests.apps.dj5plus", *INSTALLED_APPS]


MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]


TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]


# Database
# https://docs.djangoproject.com/en/3.2/ref/settings/#databases

rdbms = os.environ.get("RDBMS", "sqlite")

if rdbms == "sqlite":
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": BASE_DIR / "db.sqlite3",
            "TEST": {"NAME": BASE_DIR / "db.sqlite3"},
        }
    }
elif rdbms == "postgres":
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.postgresql",
            "NAME": os.environ.get("POSTGRES_DB", "postgres"),
            "USER": os.environ.get("POSTGRES_USER", "postgres"),
            "PASSWORD": os.environ.get("POSTGRES_PASSWORD", ""),
            "HOST": os.environ.get("POSTGRES_HOST", ""),
            "PORT": os.environ.get("POSTGRES_PORT", ""),
        }
    }


# Password validation
# https://docs.djangoproject.com/en/3.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.2/topics/i18n/

LANGUAGE_CODE = "en-us"
# LANGUAGE_CODE = 'de'


TIME_ZONE = "UTC"

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.2/howto/static-files/

STATIC_URL = "/static/"

# Default primary key field type
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"


SETTINGS_FILE = 1

ROOT_URLCONF = "tests.urls"



import typing as t

from django.core.management.base import CommandError
from django.utils.translation import gettext_lazy as _
from typer import Argument, Option

from django_typer.management import TyperCommand
from tests.apps.examples.polls.models import Question as Poll


def get_poll_from_id(poll: t.Union[str, Poll]) -> Poll:
    if isinstance(poll, Poll):
        return poll
    try:
        return Poll.objects.get(pk=int(poll))
    except Poll.DoesNotExist:
        raise CommandError(f'Poll "{poll}" does not exist')


class Command(TyperCommand, rich_markup_mode="markdown"):
    def handle(
        self,
        polls: t.Annotated[
            t.List[Poll],
            Argument(
                parser=get_poll_from_id,
                help=_("The database IDs of the poll(s) to close."),
            ),
        ],
        delete: t.Annotated[
            bool,
            Option(
                # we can also get rid of that unnecessary --no-delete flag
                "--delete",
                help=_("Delete poll instead of closing it."),
            ),
        ] = False,
    ):
        """
        :sparkles: As mentioned in the last section, helps can also be set in
        the docstring and rendered using either
        [rich](https://rich.readthedocs.io/en/stable/markup.html)
        or [markdown](https://www.markdownguide.org/) :sparkles:
        """
        for poll in polls:
            poll.opened = False
            poll.save()
            self.stdout.write(
                self.style.SUCCESS(f'Successfully closed poll "{poll.id}"')
            )
            if delete:
                poll.delete()



"""
Test that multi-level and multiple inheritance resolves correctly.
"""

from django_typer.management import TyperCommand, command, group, initialize


class Command(TyperCommand):
    """
    Inheritance1
    """

    @initialize(invoke_without_command=True)
    def init(self):
        assert isinstance(self, Command)
        return "inheritance1::init()"

    @command()
    def a(self):
        assert isinstance(self, Command)
        return "inheritance1::a()"

    @group(invoke_without_command=True)
    def g(self):
        assert isinstance(self, Command)
        return "inheritance1::g()"

    @g.command()
    def ga(self):
        assert isinstance(self, Command)
        return "inheritance1::g::ga()"

    @g.command()
    def gb(self):
        assert isinstance(self, Command)
        return "inheritance1::g::gb()"



from django_typer.management import Typer

app = Typer()


@app.callback()
def init():
    return "upstream:init"


@app.command()
def sub1():
    return "upstream:sub1"


@app.command()
def sub2():
    return "upstream:sub2"


@app.group()
def grp1():
    return "upstream:grp1"


@app.grp1.command()
def grp1_cmd1():
    return "upstream:grp1_cmd1"



from django_typer.management import TyperCommand, command, finalize
from click import get_current_context


class Command(TyperCommand, chain=True):
    suppressed_base_arguments = {
        "verbosity",
        "skip_checks",
        "version",
        "settings",
        "pythonpath",
    }

    @finalize()
    def final(self, result, **kwargs):
        assert isinstance(self, Command)
        click_params = getattr(get_current_context(silent=True), "params", {})
        assert len(kwargs) == len(click_params)
        for key, value in kwargs.items():
            assert key in click_params
            assert click_params[key] == value
        return "finalized: {} | {}".format(result, kwargs)

    @command()
    def cmd1(self, arg1: int = 1):
        return "cmd1 {}".format(arg1)

    @command()
    def cmd2(self, arg2: int):
        return "cmd2 {}".format(arg2)



from django_typer.utils import (
    get_usage_script,
    accepts_var_kwargs,
    get_win_shell,
    detect_shell,
    parse_iso_duration,
    duration_iso_string,
)
from django.test import override_settings
from django.core.management import call_command
from pathlib import Path
from shellingham import ShellDetectionFailure
import shutil
import pytest
import subprocess
import sys
import os


check_frame = Path(__file__).parent / "frame_check.py"


def test_get_usage_script():
    assert (
        get_usage_script("/root/path/to/scrapt")
        == Path("/root/path/to/scrapt").absolute()
    )


def test_accepts_var_kwargs():
    def func1(a, b, **kwargs): ...

    def func2(**kwargs): ...

    def func3(named=None, **kwargs): ...

    def func4(named=None): ...

    def func5(a): ...

    def func6(): ...

    assert accepts_var_kwargs(func1)
    assert accepts_var_kwargs(func2)
    assert accepts_var_kwargs(func3)
    assert not accepts_var_kwargs(func4)
    assert not accepts_var_kwargs(func5)
    assert not accepts_var_kwargs(func6)


def test_call_frame_check():
    result = subprocess.run(
        [sys.executable, str(check_frame.absolute())], text=True, capture_output=True
    )
    assert result.stdout.splitlines() == [
        "False",
        "True",
        "False",
        "True",
        "False",
        "True",
    ]


@override_settings(INSTALLED_APPS=["tests.apps.bad", "django_typer"])
def test_register_bad_command_plugin():
    with pytest.raises(ValueError):
        call_command("bad")


@pytest.mark.skipif(
    bool(shutil.which("pwsh") or shutil.which("powershell")),
    reason="Only test when pwsh is unavailable",
)
def test_get_win_shell_no_pwsh():
    with pytest.raises(ShellDetectionFailure):
        get_win_shell()


@pytest.mark.skipif(
    bool(shutil.which("pwsh") or shutil.which("powershell")),
    reason="Only test when pwsh is unavailable",
)
def test_powershell_profile_fail():
    from django_typer.shells.powershell import PowerShellComplete

    with pytest.raises(Exception):
        PowerShellComplete().get_user_profile()


def test_detection_failure_no_env():
    shell = os.environ.pop("SHELL", "")
    try:
        with pytest.raises(ShellDetectionFailure):
            detect_shell(max_depth=0)
    finally:
        os.environ["SHELL"] = shell


def test_detection_env_fallback():
    shell = os.environ.pop("SHELL", "")
    os.environ["SHELL"] = "/bin/bash"
    try:
        assert detect_shell(max_depth=0)[0] == "bash"
    finally:
        os.environ["SHELL"] = shell


def test_detect_shell():
    assert detect_shell(max_depth=256)


def test_parse_iso_duration():
    from datetime import timedelta

    for duration in [
        timedelta(days=3, hours=4, minutes=30, seconds=15, microseconds=123456),
        timedelta(days=1, hours=12, minutes=0, seconds=0),
        timedelta(days=0, hours=23, minutes=45, seconds=30),
        timedelta(days=5, hours=0, minutes=15, seconds=5, microseconds=987654),
        timedelta(days=2, hours=8, minutes=0, seconds=0),
        -timedelta(days=3, hours=4, minutes=30, seconds=15, microseconds=123456),
        -timedelta(days=1, hours=12, minutes=0, seconds=0),
        -timedelta(days=2, hours=20, minutes=10, seconds=30),
        -timedelta(days=5, hours=6, minutes=0, seconds=50, microseconds=123000),
        -timedelta(days=10, hours=5, minutes=55, seconds=5),
        timedelta(),
        timedelta(days=1),
        timedelta(hours=2),
        timedelta(minutes=3),
        timedelta(seconds=4),
        timedelta(microseconds=5),
    ]:
        assert parse_iso_duration(duration_iso_string(duration)) == (duration, None)

    assert parse_iso_duration("") == (timedelta(), None)
    assert parse_iso_duration("-") == (-timedelta(), None)
    assert parse_iso_duration("+") == (timedelta(), None)

    with pytest.raises(ValueError):
        parse_iso_duration("?")

    with pytest.raises(ValueError):
        parse_iso_duration("=")

    with pytest.raises(ValueError):
        parse_iso_duration("P10DX1S")

    with pytest.raises(ValueError):
        parse_iso_duration("P10DT5H.43S")

    assert parse_iso_duration("-P2DT12H") == (-timedelta(days=2, hours=12), None)
    assert parse_iso_duration("P1") == (timedelta(), "1")
    assert parse_iso_duration("PT1") == (timedelta(), "1")
    assert parse_iso_duration("P2DT2H4") == (timedelta(days=2, hours=2), "4")
    assert parse_iso_duration("P2DT2H4M5") == (
        timedelta(days=2, hours=2, minutes=4),
        "5",
    )
    assert parse_iso_duration("P2DT2H4M5.") == (
        timedelta(days=2, hours=2, minutes=4, seconds=5),
        None,
    )

    # microseconds are weird
    assert parse_iso_duration("P2DT2H4M5.000123") == (
        timedelta(days=2, hours=2, minutes=4, seconds=5, microseconds=123),
        None,
    )
    assert parse_iso_duration("P2DT2H4M5.123456") == (
        timedelta(days=2, hours=2, minutes=4, seconds=5, microseconds=123456),
        None,
    )

    assert parse_iso_duration("P2DT2H4M5.000123S") == (
        timedelta(days=2, hours=2, minutes=4, seconds=5, microseconds=123),
        None,
    )
    assert parse_iso_duration("P2DT2H4M5.123456S") == (
        timedelta(days=2, hours=2, minutes=4, seconds=5, microseconds=123456),
        None,
    )

    assert parse_iso_duration("P2DT2H4M5.00012") == (
        timedelta(days=2, hours=2, minutes=4, seconds=5),
        "00012",
    )
    assert parse_iso_duration("P2DT2H4M5.12") == (
        timedelta(days=2, hours=2, minutes=4, seconds=5),
        "12",
    )

    assert parse_iso_duration("P2DT2H4M5.00012S") == (
        timedelta(days=2, hours=2, minutes=4, seconds=5, microseconds=120),
        None,
    )
    assert parse_iso_duration("P2DT2H4M5.12S") == (
        timedelta(days=2, hours=2, minutes=4, seconds=5, microseconds=120000),
        None,
    )

    assert parse_iso_duration("P2DT5.00012S") == (
        timedelta(days=2, seconds=5, microseconds=120),
        None,
    )
    assert parse_iso_duration("P2DT5.12S") == (
        timedelta(days=2, seconds=5, microseconds=120000),
        None,
    )

    assert parse_iso_duration("P2DT5.00012") == (timedelta(days=2, seconds=5), "00012")
    assert parse_iso_duration("P2DT5.12") == (timedelta(days=2, seconds=5), "12")

    assert parse_iso_duration("P2DT5.") == (timedelta(days=2, seconds=5), None)
    assert parse_iso_duration("P2DT5.") == (timedelta(days=2, seconds=5), None)

    assert parse_iso_duration("PT5") == (timedelta(), "5")
    assert parse_iso_duration("-PT5") == (-timedelta(), "5")



import json
from typing import List

from django_typer.management import TyperCommand, command


class Command(TyperCommand):
    help = "Test multiple sub-commands."

    @command()
    def cmd1(self, files: List[str], flag1: bool = False):
        """
        A command that takes a list of files and a flag.
        """
        assert self.__class__ is Command
        return json.dumps({"files": files, "flag1": flag1})

    @command()
    def sum(self, numbers: List[float]):
        """
        Sum the given numbers.
        """
        assert self.__class__ is Command
        return str(sum(numbers))

    @command()
    def cmd3(self):
        """
        A command with no arguments.
        """
        assert self.__class__ is Command
        return json.dumps({})



from typing import Optional

import typer

from django_typer.management import TyperCommand, command


class Command(TyperCommand):
    help = "Test typer echo/secho wrappers."

    @command()
    def echo_test(
        self,
        message: str,
        color: Optional[str] = None,
        error: Optional[bool] = False,
        nl: Optional[bool] = True,
    ):
        assert self.__class__ is Command
        if color:
            self.echo(typer.style(message, fg=color), nl=nl, err=error)
        else:
            self.echo(message, nl=nl, err=error)

    @command()
    def secho_test(
        self,
        message: str,
        color: Optional[str] = None,
        error: Optional[bool] = False,
        nl: Optional[bool] = True,
    ):
        assert self.__class__ is Command
        self.secho(message, nl=nl, err=error, fg=color)



from tests.apps.test_app.management.commands.metavar3 import (
    Command as Metavar3Command,
)


@Metavar3Command.initialize(options_metavar="{INIT OVERRIDE}")
def init():
    print("init_override()")


@Metavar3Command.command(options_metavar="(CMD OVERRIDE)")
def cmd():
    return "cmd_override()"


@Metavar3Command.group(options_metavar="[GRP OVERRIDE]")
def grp():
    return "grp_override()"



from django_typer.management import TyperCommand, command


class Command(TyperCommand):
    @command(name="default")
    def handle(self):
        assert self.__class__ is Command
        return "handle"

    @command(name="renamed")
    def subcommand1(self):
        assert self.__class__ is Command
        return "subcommand1"

    @command("renamed2")
    def subcommand2(self):
        assert self.__class__ is Command
        return "subcommand2"



from django_typer.management import TyperCommand, command, finalize


class Command(TyperCommand):
    @finalize()
    def final(self, result, **_):
        assert isinstance(self, Command)
        assert result == "handle"
        return "finalized: {}".format(result)

    def handle(self):
        return "handle"



import json
from io import StringIO

import django
from django.core.management import call_command
from django.test import TestCase

from django_typer.management import get_command
from tests.utils import run_command


class MultiTests(TestCase):
    def test_command_line(self):
        self.assertEqual(
            run_command("multi", "cmd1", "/path/one", "/path/two")[0],
            {"files": ["/path/one", "/path/two"], "flag1": False},
        )

        self.assertEqual(
            run_command("multi", "cmd1", "/path/four", "/path/three", "--flag1")[0],
            {"files": ["/path/four", "/path/three"], "flag1": True},
        )

        self.assertEqual(
            run_command("multi", "sum", "1.2", "3.5", " -12.3")[0],
            sum([1.2, 3.5, -12.3]),
        )

        self.assertEqual(run_command("multi", "cmd3")[0], {})

    def test_call_command(self):
        ret = json.loads(call_command("multi", ["cmd1", "/path/one", "/path/two"]))
        self.assertEqual(ret, {"files": ["/path/one", "/path/two"], "flag1": False})

        ret = json.loads(
            call_command("multi", ["cmd1", "/path/four", "/path/three", "--flag1"])
        )
        self.assertEqual(ret, {"files": ["/path/four", "/path/three"], "flag1": True})

        ret = json.loads(call_command("multi", ["sum", "1.2", "3.5", " -12.3"]))
        self.assertEqual(ret, sum([1.2, 3.5, -12.3]))

        ret = json.loads(call_command("multi", ["cmd3"]))
        self.assertEqual(ret, {})

    def test_call_command_stdout(self):
        out = StringIO()
        call_command("multi", ["cmd1", "/path/one", "/path/two"], stdout=out)
        self.assertEqual(
            json.loads(out.getvalue()),
            {"files": ["/path/one", "/path/two"], "flag1": False},
        )

        out = StringIO()
        call_command(
            "multi", ["cmd1", "/path/four", "/path/three", "--flag1"], stdout=out
        )
        self.assertEqual(
            json.loads(out.getvalue()),
            {"files": ["/path/four", "/path/three"], "flag1": True},
        )

        out = StringIO()
        call_command("multi", ["sum", "1.2", "3.5", " -12.3"], stdout=out)
        self.assertEqual(json.loads(out.getvalue()), sum([1.2, 3.5, -12.3]))

        out = StringIO()
        call_command("multi", ["cmd3"], stdout=out)
        self.assertEqual(json.loads(out.getvalue()), {})

    def test_get_version(self):
        self.assertEqual(
            str(run_command("multi", "--version")[0]).strip(), django.get_version()
        )
        self.assertEqual(
            str(run_command("multi", "--version", "cmd1")[0]).strip(),
            django.get_version(),
        )
        self.assertEqual(
            str(run_command("multi", "--version", "sum")[0]).strip(),
            django.get_version(),
        )
        self.assertEqual(
            str(run_command("multi", "--version", "cmd3")[0]).strip(),
            django.get_version(),
        )

    def test_call_direct(self):
        multi = get_command("multi")

        self.assertEqual(
            json.loads(multi.cmd1(["/path/one", "/path/two"])),
            {"files": ["/path/one", "/path/two"], "flag1": False},
        )

        self.assertEqual(
            json.loads(multi.cmd1(["/path/four", "/path/three"], flag1=True)),
            {"files": ["/path/four", "/path/three"], "flag1": True},
        )

        self.assertEqual(float(multi.sum([1.2, 3.5, -12.3])), sum([1.2, 3.5, -12.3]))

        self.assertEqual(json.loads(multi.cmd3()), {})



import json

from django.utils.translation import gettext_lazy as _

from .help_precedence8 import Command as BaseCommand


# if you really want to use a class docstring to override a base class help at
# higher precedence you can unset the base class help attribute by setting the
# higher precedence values to empty strings
class Command(BaseCommand, help=_("")):
    """
    Now class docstring is used!
    """

    help = ""

    def handle(self, arg1: str, arg2: str, arg3: float = 0.5, arg4: int = 1):
        assert self.__class__ is Command
        opts = {"arg1": arg1, "arg2": arg2, "arg3": arg3, "arg4": arg4}
        return json.dumps(opts)



import contextlib
from io import StringIO

from django.core.management import call_command
from django.test import TestCase
from django.utils import timezone
import contextlib

from django_typer.management import get_command
from tests.apps.examples.polls.models import Question

SHELLS = [("zsh", True), ("bash", False), ("pwsh", True), ("fish", True)]


class TestPollExample(TestCase):
    q1 = None
    q2 = None
    q3 = None

    databases = {"default"}

    typer = ""

    def setUp(self):
        self.q1 = Question.objects.create(
            question_text="Is Putin a war criminal?",
            pub_date=timezone.now(),
        )
        self.q2 = Question.objects.create(
            question_text="Is Bibi a war criminal?",
            pub_date=timezone.now(),
        )
        self.q3 = Question.objects.create(
            question_text="Is Xi a Pooh Bear?",
            pub_date=timezone.now(),
        )
        super().setUp()

    def tearDown(self):
        Question.objects.all().delete()
        super().tearDown()

    def test_poll_complete(self):
        # result = run_command("shellcompletion", "complete", "./manage.py closepoll ")

        for shell, has_help in SHELLS:
            result1 = StringIO()
            with contextlib.redirect_stdout(result1):
                call_command(
                    "shellcompletion",
                    "complete",
                    shell=shell,
                    command=f"closepoll{self.typer} ",
                )
            result2 = StringIO()
            with contextlib.redirect_stdout(result2):
                call_command(
                    "shellcompletion",
                    "complete",
                    shell=shell,
                    command=f"./manage.py closepoll{self.typer} ",
                )

            result = result1.getvalue()
            self.assertEqual(result, result2.getvalue())
            for q in [self.q1, self.q2, self.q3]:
                self.assertTrue(str(q.id) in result)
                if has_help:
                    self.assertTrue(q.question_text in result)

    def test_tutorial1(self):
        with contextlib.redirect_stdout(StringIO()) as output:
            call_command(f"closepoll_t1{self.typer}", str(self.q2.id))
            result = output.getvalue()
        self.assertEqual(result.strip(), f'Successfully closed poll "{self.q2.id}"')

    def test_tutorial2(self):
        with contextlib.redirect_stdout(StringIO()) as output:
            call_command(f"closepoll_t2{self.typer}", str(self.q2.id))
            result = output.getvalue()
        self.assertEqual(result.strip(), f'Successfully closed poll "{self.q2.id}"')

    def test_tutorial_parser(self):
        with contextlib.redirect_stdout(StringIO()) as output:
            call_command(f"closepoll_t3{self.typer}", str(self.q1.id))
            result = output.getvalue()
        self.assertEqual(result.strip(), f'Successfully closed poll "{self.q1.id}"')

    def test_tutorial_parser_cmd(self):
        log = StringIO()
        call_command(f"closepoll_t3{self.typer}", str(self.q1.id), stdout=log)
        cmd = get_command(f"closepoll_t3{self.typer}", stdout=log)
        cmd([self.q1])
        cmd(polls=[self.q1])
        # these don't work, maybe revisit in future?
        # cmd([str(self.q1.id)])
        # cmd([self.q1.id])
        self.assertEqual(log.getvalue().count("Successfully"), 3)

    def test_tutorial_modelobjparser_cmd(self):
        log = StringIO()
        call_command(f"closepoll_t6{self.typer}", str(self.q1.id), stdout=log)
        cmd = get_command(f"closepoll_t6{self.typer}", stdout=log)
        cmd([self.q1])
        cmd(polls=[self.q1])
        self.assertEqual(log.getvalue().count("Successfully"), 3)

    def test_poll_ex(self):
        with contextlib.redirect_stdout(StringIO()) as output:
            call_command(f"closepoll{self.typer}", str(self.q2.id))
            result = output.getvalue()
        self.assertTrue("Successfully closed poll" in result)


class TestPollExampleTyper(TestPollExample):
    typer = "_typer"



import typing as t
from django_typer.management import TyperCommand, command, finalize


# chain=True allows multiple subroutines to be called from the command line
class Command(TyperCommand, chain=True):
    @finalize()
    def to_csv(self, results: t.List[str]):
        return ", ".join(results)

    @command()
    def cmd1(self):
        return "result1"

    @command()
    def cmd2(self):
        return "result2"



import collections.abc as cabc
import io
import sys
import typing as t
from abc import abstractmethod
from functools import cached_property
from importlib.resources import files
from pathlib import Path

from click.core import Command as ClickCommand
from click.shell_completion import CompletionItem, ShellComplete, add_completion_class
from django.template import Context, Engine
from django.template.backends.django import Template as DjangoTemplate
from django.template.base import Template as BaseTemplate
from django.template.loader import TemplateDoesNotExist, get_template
from django.utils.translation import gettext as _

__all__ = ["DjangoTyperShellCompleter", "register_completion_class"]

if t.TYPE_CHECKING:  # pragma: no cover
    from django_typer.management.commands.shellcompletion import (
        Command as ShellCompletion,
    )


class DjangoTyperShellCompleter(ShellComplete):
    """
    An extension to the click shell completion classes that provides a Django specific
    shell completion implementation. If you wish to support a new shell, you must
    derive from this class, implement all of the abstract methods, and register the
    class with
    :func:`~django_typer.management.commands.shellcompletion.register_completion_class`

    :param cli: The click command object to complete
    :param ctx_args: Additional context arguments to pass to the typer/click context of
        the management command being completed
    :param prog_name: The name of the command to complete
    :param complete_var: The name of the environment variable to pass the completion
        string (unused)
    :param command: The Django shellcompletion command
    :param command_str: The command string to complete
    :param command_args: The command arguments to complete
    :param template: The name of the shell completion script template
    :param color: Allow or disallow color and formatting in the completion output
    """

    template: str
    """
    The name of the shell completion function script template.
    """

    color: bool = False
    """
    By default, allow or disallow color and formatting in the completion output.
    """

    supports_scripts: bool = False
    """
    Does the shell support completions for uninstalled scripts? (i.e. not on the path)
    """

    complete_var: str = ""

    command: "ShellCompletion"
    command_str: str
    command_args: t.List[str]

    console = None  # type: ignore[var-annotated]

    rich_console = None  # type: ignore[var-annotated]
    console_buffer: io.StringIO

    color_default: bool = True

    def __init__(
        self,
        cli: t.Optional[ClickCommand] = None,
        ctx_args: cabc.MutableMapping[str, t.Any] = {},
        prog_name: str = "",
        complete_var: str = "",
        command: t.Optional["ShellCompletion"] = None,
        command_str: t.Optional[str] = None,
        command_args: t.Optional[t.List[str]] = None,
        template: t.Optional[str] = None,
        color: t.Optional[bool] = None,
        color_default: bool = color_default,
        **kwargs,
    ):
        # we don't always need the initialization parameters during completion
        self.prog_name = prog_name
        if command:
            self.command = command
        if command_str is not None:
            self.command_str = command_str
        if command_args is not None:
            self.command_args = command_args
        if template is not None:
            self.template = template
        if color is not None:
            self.color = color
        self.color_default = color_default

        self.console_buffer = io.StringIO()
        try:
            from rich.console import Console

            self.console = Console(
                # do not disable color output if not explicitly disabled
                color_system="auto" if self.color_default or self.color else None,
                force_terminal=True,
                file=command.stdout if command else None,  # type: ignore[arg-type]
                stderr=command.stderr if command else None,  # type: ignore[arg-type]
            )
            self.rich_console = Console(
                color_system="auto" if self.color else None,
                force_terminal=True,
                file=self.console_buffer,
            )
        except ImportError:
            pass

        if cli:
            super().__init__(
                cli=cli,
                ctx_args=ctx_args,
                complete_var=complete_var,
                prog_name=prog_name,
                **kwargs,
            )

    @property
    def source_template(self) -> str:  # type: ignore
        """
        The contents of the shell's completion script template.
        """
        return Path(self.load_template().origin.name).read_text()

    def get_completions(
        self, args: t.List[str], incomplete: str
    ) -> t.List[CompletionItem]:
        """
        Get the completions for the current command string and incomplete string.
        """
        if self.command.fallback:
            return self.command.fallback(args, incomplete)
        return super().get_completions(args[1:], incomplete)

    def get_completion_args(self) -> t.Tuple[t.List[str], str]:
        """
        Return the list of completion arguments and the incomplete string.
        """
        cwords = self.command_args
        if self.command_str and self.command_str[-1].isspace():
            # if the command string ends with a space, the incomplete string is empty
            cwords.append("")
        return (
            cwords[:-1],
            cwords[-1] if cwords else "",
        )

    def source_vars(self) -> t.Dict[str, t.Any]:
        """
        This returns the context that will be used to render the completion script
        template.

        From the base class we inherit the following variables:

        * **complete_func**: the name to use for the shell's completion function
        * **complete_var**: unused because we do not use environment variables
            to pass the completion string
        * **prog_name**: the name of the command to complete

        We add the following variables:

        * **manage_script**: the manage script object - will be either a string or a
          Path, if it is a Path it will be absolute and this indicates that the script
          is not installed on the path
        * **manage_script_name**: the name of the Django manage script
        * **python**: the path to the python interpreter that is running the
          shellcompletion command
        * **django_command**: the name of the Django shellcompletion command - you may
          change this to something other than 'shellcompletion' to provide custom
          complete logic
        * **color**: the color flag to pass to shellcompletion i.e. --(force|no)-color
        * **fallback**: the fallback option to pass to ``shellcompletion complete``
        * **is_installed**: whether or not the manage script is a command on the path
        * **shell**: the name of the shell
        """
        return {
            **super().source_vars(),
            "manage_script": self.command.manage_script,
            "manage_script_name": self.command.manage_script_name,
            "python": sys.executable,
            "django_command": self.command.__module__.split(".")[-1],
            "color": "--force-color"
            if self.color
            else "--no-color"
            if self.command.force_color
            else "",
            "fallback": f"--fallback={self.command.fallback_import}"
            if self.command.fallback
            else "",
            "is_installed": self.is_installed,
            "shell": self.name,
        }

    @cached_property
    def is_installed(self) -> bool:
        """
        Whether or not the manage script is a command on the path.
        """
        return not isinstance(self.command.manage_script, Path)

    def load_template(self) -> t.Union[BaseTemplate, DjangoTemplate]:
        """
        Return a compiled Template object for the completion script template.
        """
        try:
            return get_template(self.template)  # type: ignore
        except TemplateDoesNotExist:
            # handle case where templating is not configured to find our default
            # templates
            return Engine(
                dirs=[str(files("django_typer").joinpath("templates"))],
                libraries={
                    "default": "django.template.defaulttags",
                    "filter": "django.template.defaultfilters",
                },
            ).get_template(self.template)

    def source(self) -> str:
        """
        Render the completion script template to a string.
        """
        try:
            return self.load_template().render(self.source_vars())  # type: ignore
        except (AttributeError, TypeError, ValueError):
            # it is annoying that get_template() and DjangoEngine.get_template() return
            # different interfaces
            return self.load_template().render(Context(self.source_vars()))  # type: ignore

    @abstractmethod
    def install(self, prompt: bool = True) -> t.List[Path]:
        """
        Deriving classes must implement this method to install the completion script.

        This method should return the path to the installed script.

        :param prompt: If True, prompt the user for confirmation before installing
            the completion script.
        :return: The paths that were created or edited or None if the user declined
            installation or no changes were made.
        """

    @abstractmethod
    def uninstall(self):
        """
        Deriving classes must implement this method to uninstall the completion script.
        """

    @abstractmethod
    def get_user_profile(self) -> Path:
        """
        Most shells have a profile script that is sourced when the interactive shell
        starts. Deriving classes should implement this method to return the location
        of that script.
        """

    def prompt(
        self,
        source: str,
        file: Path,
        prompt: bool = True,
        start_line: int = 0,
    ) -> bool:
        """
        Prompt the user for confirmation before editing the file with the given
        source edits.

        :param source: The source string that will be written to the file.
        :param file: The path of the file that will be created or edited.
        :param prompt: Prompt toggle, if False, will not prompt the user and return True
        :param start_line: The line number the edit will start at.
        :return: True if the user confirmed the edit, False otherwise.
        """
        if not prompt:
            return True

        prompt_text = (
            _("Append the above contents to {file}?").format(file=file)
            if start_line
            else _("Create {file} with the above contents?").format(file=file)
        )
        if self.console:
            from rich.prompt import Confirm
            from rich.syntax import Syntax

            syntax = Syntax(
                source,
                self.name,
                theme="monokai",
                start_line=start_line,
                line_numbers=False,
            )
            self.console.print(syntax)
            return Confirm.ask(prompt_text, console=self.console)
        else:
            print(source)
            return input(prompt_text + " [y/N] ").lower() in {"y", "yes"}

    def process_rich_text(self, text: str) -> str:
        """
        Removes rich text markup from a string if color is disabled, otherwise it
        will render the rich markup to ansi control codes. If rich is not installed,
        none of this happens and the markup will be passed through as is.
        """
        if self.rich_console:
            if self.color:
                self.console_buffer.truncate(0)
                self.console_buffer.seek(0)
                self.rich_console.print(text, end="")
                return self.console_buffer.getvalue()
            else:
                return "".join(
                    segment.text for segment in self.rich_console.render(text)
                ).rstrip("\n")
        return text


_completers: t.Dict[str, t.Type[DjangoTyperShellCompleter]] = {}


def register_completion_class(cls: t.Type[DjangoTyperShellCompleter]) -> None:
    """
    Register a shell completion class for use with the Django shellcompletion command.
    """
    _completers[cls.name] = cls
    add_completion_class(cls)



import sys
import typing as t

from django.conf import settings

from django_typer.management import TyperCommand, initialize


COMMAND_TMPL = """
def {routine}(self, {flag_args}):
    flags = []
    {add_flags}
    return self.run("{routine}", flags)
"""


class Command(TyperCommand):
    @initialize()
    def init(self):
        assert self.__class__ is Command

    def run(self, routine: str, flags: t.Optional[t.List[str]]):
        assert self.__class__ is Command
        return routine, flags


for routine, flags in getattr(settings, "ROUTINES", {}).items():
    flag_args = ", ".join([f"{flag}: bool=False" for flag in flags])
    add_flags = ""
    for flag in flags:
        add_flags += f'\n    if {flag}: flags.append("{flag}")'
    exec(COMMAND_TMPL.format(routine=routine, flag_args=flag_args, add_flags=add_flags))
    Command.command()(locals()[routine])



from django_typer.management import Typer

# use native Typer interface to achieve the same result


def to_csv(results, **_):
    return ", ".join(results)


def to_upper_csv(results, **_):
    return ", ".join([result.upper() for result in results])


app = Typer(result_callback=to_csv, chain=True)

grp = Typer(result_callback=to_upper_csv, chain=True)
app.add_typer(grp, name="grp")


@app.command()
def cmd1():
    return "result1"


@app.command()
def cmd2():
    return "result2"


@grp.command()
def cmd3():
    return "result3"


@grp.command()
def cmd4():
    return "result4"



from django_typer.management import TyperCommand


class Command(TyperCommand):
    help = "Constructor tests"

    def handle(self):
        assert self.__class__ is Command
        self.stderr.write(
            f"err\nno_color={self.no_color}\nforce_color={self.force_color}\n"
        )
        self.stdout.write(
            f"out\nno_color={self.no_color}\nforce_color={self.force_color}\n"
        )



from django.test import TestCase

from django_typer.management import get_command
from django_typer.utils import get_current_command


class TestGetCommand(TestCase):
    def test_get_command(self):
        from tests.apps.test_app.management.commands.basic import (
            Command as Basic,
        )

        basic = get_command("basic")
        assert basic.__class__ == Basic

        from tests.apps.test_app.management.commands.multi import (
            Command as Multi,
        )

        multi = get_command("multi")
        assert multi.__class__ == Multi
        cmd1 = get_command("multi", "cmd1")
        assert cmd1.__func__ is multi.cmd1.__func__
        sum = get_command("multi", "sum")
        assert sum.__func__ is multi.sum.__func__
        cmd3 = get_command("multi", "cmd3")
        assert cmd3.__func__ is multi.cmd3.__func__

        from tests.apps.test_app.management.commands.callback1 import (
            Command as Callback1,
        )

        callback1 = get_command("callback1")
        assert callback1.__class__ == Callback1

        # callbacks are not commands
        with self.assertRaises(LookupError):
            get_command("callback1", "init")


def test_get_current_command_returns_none():
    assert get_current_command() is None



from django_typer.management import TyperCommand, initialize, group, finalize
from click import get_current_context


class Command(TyperCommand):
    init_opt: bool
    direct_call: bool = False

    g1_opt: bool
    g2_opt: bool

    common_kwargs = {
        "force_color",
        "no_color",
        "skip_checks",
        "version",
        "settings",
        "pythonpath",
        "traceback",
    }

    @initialize()
    def init(
        self,
        init_opt: bool = False,
        direct_call: bool = direct_call,
        call_command: bool = False,
    ):
        self.init_opt = init_opt
        self.direct_call = direct_call
        return f"root_init | init_opt={init_opt}"

    @finalize()
    def root_final(self, result, init_opt: bool, call_command: bool = False, **kwargs):
        assert isinstance(self, Command)
        if not self.direct_call:
            click_params = getattr(get_current_context(silent=True), "params", {})
            assert len(kwargs) == len(click_params) - 2
            for key, value in kwargs.items():
                assert key in click_params
                assert click_params[key] == value

        if not call_command and not self.direct_call:
            assert "g1_opt" not in kwargs
            assert "g2_opt" not in kwargs
        return f"root_final: {result} | init_opt={init_opt}"

    @group(chain=True)
    def grp1(self, g1_opt: bool = False):
        self.g1_opt = g1_opt
        return "grp1"

    @group(chain=True)
    def grp2(self, g2_opt: bool = True):
        self.g2_opt = g2_opt
        return "grp2"

    @grp1.command()
    def cmd1(self):
        return "cmd1"

    @grp1.command()
    def cmd2(self):
        return "cmd2"

    @grp2.command()
    def cmd3(self):
        return "cmd3"

    @grp2.command()
    @staticmethod
    def cmd4():
        return "cmd4"

    @grp1.finalize()
    def grp1_final(self, result, **kwargs):
        assert isinstance(self, Command)
        assert "cmd1" in result or "cmd2" in result
        assert "cmd3" not in result
        assert "cmd4" not in result
        if not self.direct_call:
            assert "g1_opt" in kwargs
            assert kwargs["g1_opt"] == self.g1_opt
            if not self.direct_call:
                for key in Command.common_kwargs:
                    assert key in kwargs
            assert "g2_opt" not in kwargs
            assert "init_opt" in kwargs
        return f"grp1_final: {result} | g1_opt={kwargs.get('g1_opt', None)}"

    @grp2.finalize()
    @staticmethod
    def grp2_final(result, **kwargs):
        assert "cmd3" in result or "cmd4" in result
        assert "cmd1" not in result
        assert "cmd2" not in result

        try:
            assert kwargs["gp2_opt"] == get_current_context().django_command.g1_opt
            for key in Command.common_kwargs:
                assert key in kwargs

            assert "g1_opt" not in kwargs
            assert "init_opt" in kwargs
        except Exception:
            pass

        return f"grp2_final: {result} | g2_opt={kwargs.get('g2_opt', None)}"



from django_typer.management import TyperCommand, command, finalize
from click import get_current_context


class Command(TyperCommand, chain=True):
    suppressed_base_arguments = {
        "verbosity",
        "skip_checks",
        "version",
        "settings",
        "pythonpath",
    }

    @finalize()
    def final(
        self, result, force_color=None, no_color=None, traceback=None, show_locals=None
    ):
        assert isinstance(self, Command)
        try:
            click_params = getattr(get_current_context(silent=False), "params", {})
            assert "force_color" in click_params
            assert click_params["force_color"] == force_color
            assert "no_color" in click_params
            assert click_params["no_color"] == no_color
            assert "traceback" in click_params
            assert click_params["traceback"] == traceback
            assert "show_locals" in click_params
            assert click_params["show_locals"] == traceback
        except Exception:
            pass
        return "finalized: {} | {}".format(
            result,
            {
                "force_color": force_color,
                "no_color": no_color,
                "traceback": traceback,
                "show_locals": show_locals,
            },
        )

    @command()
    def cmd1(self, arg1: int = 1):
        return "cmd1 {}".format(arg1)

    @command()
    def cmd2(self, arg2: int):
        return "cmd2 {}".format(arg2)



import contextlib
from io import StringIO

from django.core.management import call_command
from django.test import TestCase

from django_typer.management import get_command
from tests.utils import run_command


class TestResultCallback(TestCase):
    def test_result_callback1(self):
        out = StringIO()
        call_command("result_callback1", "cmd", stdout=out)
        self.assertEqual(out.getvalue(), "finalized\n")

    def test_result_callback2(self):
        out = StringIO()
        call_command("result_callback2", "cmd1", "cmd2", stdout=out)
        self.assertEqual(out.getvalue(), "finalize\n")

    def test_result_callback3(self):
        out = StringIO()
        call_command("result_callback3", "cmd", stdout=out)
        self.assertEqual(out.getvalue(), "finalize\n")



from .interference_ext import Command as InterferenceExt


class Command(InterferenceExt):
    help = "Inherited and extend interference - test classmethod initialize()"

    @InterferenceExt.initialize(invoke_without_command=True)
    def init(self, arg: int = 5):
        assert self.__class__ is Command
        return f"test_app2::interference_ext_init::init({arg})"



import typer
from django.test import TestCase

from django_typer.management import (
    Typer,
    command,
    get_command,
    group,
    initialize,
)
from tests.utils import get_named_arguments, get_named_defaults


class InterfaceTests(TestCase):
    """
    Make sure the django_typer decorator interfaces match the
    typer decorator interfaces. We don't simply pass variadic arguments
    to the typer decorator because we want the IDE to offer auto complete
    suggestions. This is a "developer experience" concession

    Include some other interface tests designed to test compatibility between
    the overrides and what the base class expects
    """

    def compare_defaults(
        self, params, dt_function, typer_function, require_placeholder_match=True
    ):
        if "cls" in params:
            params.remove(
                "cls"
            )  # cls default is always purposefully overridden in django-typer interfaces
        # sanity
        self.assertGreater(len(params), 0)
        dt_defaults = get_named_defaults(dt_function)
        typer_defaults = get_named_defaults(typer_function)
        count = 0
        for param in params:
            dt_default = dt_defaults[param]
            ty_default = typer_defaults[param]
            if param == "rich_markup_mode":
                from typer.core import DEFAULT_MARKUP_MODE

                ty_default = DEFAULT_MARKUP_MODE
            err_msg = f"param {param} does not match between {dt_function}::{dt_default} and {typer_function}::{ty_default}"
            if isinstance(dt_default, typer.models.DefaultPlaceholder):
                if require_placeholder_match:
                    self.assertIsInstance(
                        ty_default, typer.models.DefaultPlaceholder, msg=err_msg
                    )
                dt_default = dt_default.value
            if isinstance(ty_default, typer.models.DefaultPlaceholder):
                # this will always fail out
                if require_placeholder_match:
                    self.assertIsInstance(
                        dt_defaults[param], typer.models.DefaultPlaceholder, msg=err_msg
                    )
                ty_default = ty_default.value
            self.assertEqual(dt_default, ty_default, msg=err_msg)
            count += 1
        return count

    def test_typer_command_interface_matches(self):
        dt_params = set(get_named_arguments(Typer.command))
        typer_params = set(get_named_arguments(typer.Typer.command))

        self.assertFalse(dt_params.symmetric_difference(typer_params))
        self.assertEqual(
            self.compare_defaults(dt_params, Typer.command, typer.Typer.command),
            len(dt_params),
        )

    def test_typer_callback_interface_matches(self):
        dt_params = set(get_named_arguments(Typer.callback))
        typer_params = set(get_named_arguments(typer.Typer.callback))

        self.assertFalse(dt_params.symmetric_difference(typer_params))
        self.assertEqual(
            self.compare_defaults(dt_params, Typer.callback, typer.Typer.callback),
            len(dt_params),
        )

    def test_typer_initialize_interface_matches(self):
        dt_params = set(get_named_arguments(Typer.initialize))
        typer_params = set(get_named_arguments(typer.Typer.callback))

        self.assertFalse(dt_params.symmetric_difference(typer_params))
        self.assertEqual(
            self.compare_defaults(dt_params, Typer.initialize, typer.Typer.callback),
            len(dt_params),
        )

        self.assertEqual(Typer.initialize, Typer.callback)

    def test_typer_add_typer_interface_matches(self):
        dt_params = set(get_named_arguments(Typer.add_typer))
        typer_params = set(get_named_arguments(typer.Typer.add_typer))

        self.assertFalse(dt_params.symmetric_difference(typer_params))
        self.assertEqual(
            self.compare_defaults(dt_params, Typer.add_typer, typer.Typer.add_typer),
            len(dt_params),
        )

    def test_typer_init_interface_matches(self):
        dt_params = set(get_named_arguments(Typer.__init__))
        typer_params = set(get_named_arguments(typer.Typer.__init__))
        dt_params.remove("django_command")
        dt_params.remove("parent")
        self.assertFalse(dt_params.symmetric_difference(typer_params))
        dt_params.remove("pretty_exceptions_show_locals")
        self.assertEqual(
            self.compare_defaults(dt_params, Typer.__init__, typer.Typer.__init__),
            len(dt_params),
        )

    def test_command_interface_matches(self):
        command_params = set(get_named_arguments(command))
        typer_params = set(get_named_arguments(typer.Typer.command))

        self.assertFalse(command_params.symmetric_difference(typer_params))
        self.assertEqual(
            self.compare_defaults(command_params, command, typer.Typer.command),
            len(command_params),
        )

    def test_group_interface_matches(self):
        command_params = set(get_named_arguments(group))
        typer_params = set(get_named_arguments(typer.Typer.add_typer))
        typer_params.remove("callback")

        command_params.remove("options_metavar")
        typer_params.remove("options_metavar")

        self.assertFalse(command_params.symmetric_difference(typer_params))
        self.assertEqual(
            self.compare_defaults(command_params, group, typer.Typer.add_typer),
            len(command_params),
        )

    def test_initialize_interface_matches(self):
        from django_typer.management import callback

        initialize_params = set(get_named_arguments(initialize))
        typer_params = set(get_named_arguments(typer.Typer.callback))

        self.assertFalse(initialize_params.symmetric_difference(typer_params))
        self.assertEqual(
            self.compare_defaults(initialize_params, initialize, typer.Typer.callback),
            len(initialize_params),
        )

        self.assertTrue(initialize is callback)

    def test_typercommandmeta_interface_matches(self):
        from django_typer.management import TyperCommandMeta

        typer_command_params = set(get_named_arguments(TyperCommandMeta.__new__))
        typer_params = set(get_named_arguments(typer.Typer.__init__))
        typer_params.remove("add_completion")
        self.assertFalse(typer_command_params.symmetric_difference(typer_params))

        typer_command_params.remove("pretty_exceptions_enable")
        typer_command_params.remove("pretty_exceptions_show_locals")
        typer_command_params.remove("pretty_exceptions_short")
        self.assertEqual(
            self.compare_defaults(
                typer_command_params, TyperCommandMeta.__new__, typer.Typer.__init__
            ),
            len(typer_command_params),
        )
        self.assertEqual(
            self.compare_defaults(
                [
                    "pretty_exceptions_enable",
                    # "pretty_exceptions_show_locals",
                    "pretty_exceptions_short",
                ],
                TyperCommandMeta.__new__,
                typer.Typer.__init__,
                require_placeholder_match=False,
            ),
            2,
        )

    def test_typer_group_interface_matches(self):
        from django_typer.management import Typer

        typer_command_params = set(get_named_arguments(Typer.group))
        typer_params = set(get_named_arguments(typer.Typer.add_typer))
        typer_params.remove("callback")
        self.assertFalse(typer_command_params.symmetric_difference(typer_params))
        self.assertEqual(
            self.compare_defaults(
                typer_command_params, Typer.group, typer.Typer.add_typer
            ),
            len(typer_command_params),
        )

    def test_base_class_command_interface_matches(self):
        from django_typer.management import TyperCommand

        command_params = set(get_named_arguments(TyperCommand.command))
        typer_params = set(get_named_arguments(typer.Typer.command))

        self.assertFalse(command_params.symmetric_difference(typer_params))
        self.assertEqual(
            self.compare_defaults(
                command_params, TyperCommand.command, typer.Typer.command
            ),
            len(command_params),
        )

    def test_base_class_initialize_interface_matches(self):
        from django_typer.management import TyperCommand

        command_params = set(get_named_arguments(TyperCommand.initialize))
        typer_params = set(get_named_arguments(typer.Typer.callback))

        self.assertFalse(command_params.symmetric_difference(typer_params))
        self.assertEqual(
            self.compare_defaults(
                command_params, TyperCommand.initialize, typer.Typer.callback
            ),
            len(command_params),
        )

        self.assertEqual(TyperCommand.initialize, TyperCommand.callback)

    def test_base_class_group_interface_matches(self):
        from django_typer.management import TyperCommand

        command_params = set(get_named_arguments(TyperCommand.group))
        typer_params = set(get_named_arguments(typer.Typer.add_typer))
        typer_params.remove("callback")
        self.assertFalse(command_params.symmetric_difference(typer_params))
        self.assertEqual(
            self.compare_defaults(
                command_params, TyperCommand.group, typer.Typer.add_typer
            ),
            len(command_params),
        )

    def test_action_nargs(self):
        # unclear if nargs is even necessary - no other test seems to exercise it, leaving in for
        # base class compat reasons
        from tests.utils import rich_installed

        self.assertEqual(
            get_command("basic")
            .create_parser("./manage.py", "basic")
            ._actions[0]
            .nargs,
            1,
        )
        self.assertEqual(
            get_command("completion")
            .create_parser("./manage.py", "completion")
            ._actions[0]
            .nargs,
            -1,
        )
        multi_parser = get_command("multi").create_parser("./manage.py", "multi")
        self.assertEqual(
            multi_parser._actions[8 if rich_installed else 7].param.name, "files"
        )
        self.assertEqual(multi_parser._actions[8 if rich_installed else 7].nargs, -1)
        self.assertEqual(
            multi_parser._actions[9 if rich_installed else 8].param.name, "flag1"
        )
        self.assertEqual(multi_parser._actions[9 if rich_installed else 8].nargs, 0)

    def test_cmd_getattr(self):
        from django_typer.management import TyperCommand
        from tests.apps.test_app.management.commands.groups import (
            Command as Groups,
        )

        self.assertIsInstance(Groups.math.divide, typer.models.CommandInfo)
        try:
            Groups.does_not_exist
            self.assertFalse(True, "should have thrown AttributeError")
        except AttributeError:
            pass

        try:
            Groups.math.does_not_exist
            self.assertFalse(True, "should have thrown AttributeError")
        except AttributeError:
            pass

        try:
            TyperCommand.does_not_exist
            self.assertFalse(True, "should have thrown AttributeError")
        except AttributeError:
            pass

    def test_proxied_property(self):
        from django_typer.management import BoundProxy
        from tests.apps.test_app.management.commands.basic import (
            Command as Basic,
        )

        basic = get_command("basic", Basic)
        self.assertIsInstance(basic.typer_app, BoundProxy)
        self.assertTrue(basic.typer_app.django_command is Basic)
        self.assertTrue(basic.typer_app.proxied.django_command is Basic)

    def test_attribute_access_in_definition(self):
        from tests.apps.test_app.management.commands.groups import (
            Command as Groups,
        )

        with self.assertRaises(AttributeError):

            class Command(Groups):
                @Groups.maths.command()
                def sqrt(self):
                    pass

    def test_attribute_access_outside_of_definition(self):
        from django_typer.management import TyperCommand, TyperCommandMeta
        from tests.apps.test_app.management.commands.native import app

        with self.assertRaises(AttributeError):

            @app.maths.group()
            def sqrt():
                pass

        with self.assertRaises(AttributeError):
            TyperCommand.maths

        with self.assertRaises(AttributeError):
            TyperCommandMeta.maths

        with self.assertRaises(AttributeError):
            TyperCommandMeta("dummy", tuple(), {}).maths



from django_typer.management import TyperCommand, group


class Command(TyperCommand):
    @group()
    def g0():
        pass

    @group()
    def g1(self):
        assert self.__class__ is Command

    @g0.group(invoke_without_command=True)
    def l2(self, arg: int):
        assert self.__class__ is Command
        return f"g0:l2({arg})"

    @g1.group(invoke_without_command=True)
    def l2(arg: str):
        return f"g1:l2({arg})"

    @g0.l2.command()
    def cmd():
        return "g0:l2:cmd()"

    @g1.l2.command()
    def cmd(self):
        assert self.__class__ is Command
        return "g1:l2:cmd()"



import typing as t

from django.conf import settings


def traceback_config() -> t.Dict[str, t.Any]:
    """
    Fetch the rich traceback installation parameters from our settings. By default
    rich tracebacks are on with show_locals = True. If the config is set to False
    or None rich tracebacks will not be installed even if the library is present.

    This allows us to have a common traceback configuration for all commands. If rich
    tracebacks are managed separately this setting can also be switched off.
    """
    default = {"show_locals": False}
    cfg = getattr(settings, "DT_RICH_TRACEBACK_CONFIG", default) or default
    if cfg is True:
        return default
    return cfg


def show_locals() -> t.Optional[bool]:
    """
    Return the show_locals parameter from the rich traceback configuration.
    """
    return traceback_config().get("show_locals", None)


def use_rich_tracebacks() -> bool:
    """
    Return true if rich tracebacks should be installed, False otherwise.
    """
    from .utils import rich_installed

    cfg = getattr(settings, "DT_RICH_TRACEBACK_CONFIG", {"show_locals": False})
    return rich_installed and (
        (isinstance(cfg, dict) and not cfg.get("no_install", False)) or cfg is True
    )



from .backup_inherit import *

INSTALLED_APPS = [
    "tests.apps.examples.plugins.files2",
    "tests.apps.examples.plugins.files1",
    "tests.apps.examples.plugins.backup_files",
    *INSTALLED_APPS,
]



import inspect
import os
import typing as t
from pathlib import Path

import typer
from django.conf import settings
from django.core.management import CommandError, call_command

from django_typer.management import CommandNode, Typer
from django_typer.completers.path import directories
from django_typer.completers.db import databases as complete_db

app = Typer()

# these two lines are not necessary but will make your type checker happy
assert app.django_command
Command = app.django_command

Command.suppressed_base_arguments = {"verbosity", "skip_checks"}
Command.requires_migrations_checks = False
Command.requires_system_checks = []

databases = [alias for alias in settings.DATABASES.keys()]


@app.callback(invoke_without_command=True)
def init_or_run_all(
    self,
    # if we add a context argument Typer will provide it
    # the context is a click object that contains additional
    # information about the broader CLI invocation
    context: typer.Context,
    output_directory: t.Annotated[
        Path,
        typer.Option(
            "-o",
            "--output",
            shell_complete=directories,
            help="The directory to write backup artifacts to.",
        ),
    ] = Path(os.getcwd()),
):
    """
    Backup the website! This command groups backup routines together.
    Each routine may be run individually, but if no routine is specified,
    the default run of all routines will be executed.
    """
    self.output_directory = output_directory

    if not self.output_directory.exists():
        self.output_directory.mkdir(parents=True)

    if not self.output_directory.is_dir():
        raise CommandError(f"{self.output_directory} is not a directory.")

    # here we use the context to determine if a subcommand was invoked and
    # if it was not we run all the backup routines
    if not context.invoked_subcommand:
        for cmd in get_backup_routines(self):
            cmd()


@app.command()
def list(self):
    """
    List the default backup routines in the order they will be run.
    """
    self.echo("Default backup routines:")
    for cmd in get_backup_routines(self):
        sig = {
            name: param.default
            for name, param in inspect.signature(
                cmd.callback
            ).parameters.items()
            if not name == "self"
        }
        params = ", ".join([f"{k}={v}" for k, v in sig.items()])
        self.secho(f"  {cmd.name}({params})", fg="green")


@app.command()
def database(
    self,
    filename: t.Annotated[
        str,
        typer.Option(
            "-f",
            "--filename",
            help=(
                "The name of the file to use for the backup fixture. The "
                "filename may optionally contain a {database} formatting "
                "placeholder."
            ),
        ),
    ] = "{database}.json",
    databases: t.Annotated[
        t.Optional[t.List[str]],
        typer.Option(
            "-d",
            "--database",
            help=(
                "The name of the database(s) to backup. If not provided, "
                "all databases will be backed up."
            ),
            shell_complete=complete_db(),
        ),
    ] = databases,
):
    """
    Backup database(s) to a json fixture file.
    """
    for db in databases or self.databases:
        output = self.output_directory / filename.format(database=db)
        self.echo(f"Backing up database [{db}] to: {output}")
        call_command(
            "dumpdata",
            output=output,
            database=db,
            format="json",
        )


def get_backup_routines(command) -> t.List[CommandNode]:
    """
    Return the list of backup subcommands. This is every registered command
    except for the list command.
    """
    # fetch all the command names at the top level of our command tree,
    # except for list, which we know to not be a backup routine
    return [
        cmd
        for name, cmd in command.get_subcommand().children.items()
        if name != "list"
    ]



import sys
from io import StringIO
from pathlib import Path

from django.core.management import call_command
from django.test import TestCase, override_settings

from django_typer.management import get_command
from tests.utils import read_django_parameters, run_command


class TestDjangoParameters(TestCase):
    commands = [
        ("dj_params1", []),
        ("dj_params2", ["cmd1"]),
        ("dj_params2", ["cmd2"]),
        ("dj_params3", ["cmd1"]),
        ("dj_params3", ["cmd2"]),
        ("dj_params4", []),
    ]

    def test_settings(self):
        for cmd, args in self.commands:
            run_command(cmd, "--settings", "tests.settings.settings2", *args)
            self.assertEqual(read_django_parameters().get("settings", None), 2)

    def test_color_params(self):
        for cmd, args in self.commands:
            run_command(cmd, "--no-color", *args)
            params = read_django_parameters()
            self.assertEqual(params.get("no_color", False), True)
            self.assertEqual(params.get("no_color_attr", False), True)
            run_command(cmd, "--force-color", *args)
            params = read_django_parameters()
            self.assertEqual(params.get("no_color", True), False)
            self.assertEqual(params.get("no_color_attr", True), False)

            result = run_command(cmd, "--force-color", "--no-color", *args)
            self.assertTrue(
                "The --no-color and --force-color options can't be used together."
                in result[1]
            )

            call_command(cmd, args, no_color=True)
            params = read_django_parameters()
            self.assertEqual(params.get("no_color", False), True)
            self.assertEqual(params.get("no_color_attr", False), True)
            call_command(cmd, args, force_color=True)
            params = read_django_parameters()
            self.assertEqual(params.get("no_color", True), False)
            self.assertEqual(params.get("no_color_attr", True), False)
            with self.assertRaises(BaseException):
                call_command(cmd, args, force_color=True, no_color=True)

    def test_ctor_params(self):
        # check non-tty streams output expected constructor values and coloring
        stdout = StringIO()
        stderr = StringIO()
        cmd = get_command(
            "ctor", stdout=stdout, stderr=stderr, no_color=None, force_color=None
        )
        cmd()
        out_str = stdout.getvalue()
        err_str = stderr.getvalue()
        self.assertEqual(out_str, "out\nno_color=None\nforce_color=None\n")
        self.assertEqual(err_str, "err\nno_color=None\nforce_color=None\n")
        cmd.print_help("./manage.py", "ctor")

        # check no-color
        stdout = StringIO()
        stderr = StringIO()
        stdout.isatty = lambda: True
        stderr.isatty = lambda: True
        cmd = get_command(
            "ctor", stdout=stdout, stderr=stderr, no_color=True, force_color=False
        )
        cmd()
        self.assertEqual(stdout.getvalue(), "out\nno_color=True\nforce_color=False\n")
        self.assertEqual(stderr.getvalue(), "err\nno_color=True\nforce_color=False\n")
        cmd.print_help("./manage.py", "ctor")
        self.assertTrue("\x1b" not in stdout.getvalue())
        self.assertTrue("\x1b" not in stderr.getvalue())
        stdout.truncate(0)
        stderr.truncate(0)

        stdout.getvalue()
        stderr.getvalue()
        cmd.execute(skip_checks=False, no_color=None, force_color=None)
        out_str = stdout.getvalue()
        err_str = stderr.getvalue()
        self.assertTrue(out_str.endswith("out\nno_color=True\nforce_color=False\n"))
        self.assertTrue(err_str.endswith("err\nno_color=True\nforce_color=False\n"))

    def test_pythonpath(self):
        added = str(Path(__file__).parent.absolute())
        self.assertTrue(added not in sys.path)
        for cmd, args in self.commands:
            run_command(cmd, "--pythonpath", added, *args)
            self.assertTrue(added in read_django_parameters().get("python_path", []))

    def test_skip_checks(self):
        for cmd, args in self.commands:
            result = run_command(
                cmd,
                "--settings",
                "tests.settings.settings_fail_check",
                *args,
            )
            self.assertTrue("SystemCheckError" in result[1])
            self.assertTrue("test_app.E001" in result[1])

            result = run_command(
                cmd,
                "--skip-checks",
                "--settings",
                "tests.settings.settings_fail_check",
                *args,
            )
            self.assertFalse("SystemCheckError" in result[1])
            self.assertFalse("test_app.E001" in result[1])

    @override_settings(DJANGO_TYPER_FAIL_CHECK=True)
    def test_skip_checks_call(self):
        for cmd, args in self.commands:
            from django.core.management.base import SystemCheckError

            with self.assertRaises(SystemCheckError):
                call_command(cmd, *args, skip_checks=False)

            # when you call_command and don't supply skip_checks, it will default to True!
            call_command(cmd, *args, skip_checks=True)
            call_command(cmd, *args)

    def test_traceback(self):
        # traceback does not come into play with call_command
        for cmd, args in self.commands:
            result = run_command(cmd, *args, "--throw")[1]
            if cmd != "dj_params4":
                self.assertFalse("Traceback" in result)
            else:
                self.assertTrue("Traceback" in result)

            if cmd != "dj_params4":
                result_tb = run_command(cmd, "--traceback", *args, "--throw")[1]
                self.assertTrue("Traceback" in result_tb)
            else:
                result_tb = run_command(cmd, "--no-traceback", *args, "--throw")[1]
                self.assertFalse("Traceback" in result_tb)

    def test_verbosity(self):
        run_command("dj_params3", "cmd1")
        self.assertEqual(read_django_parameters().get("verbosity", None), 1)

        call_command("dj_params3", ["cmd1"])
        self.assertEqual(read_django_parameters().get("verbosity", None), 1)

        run_command("dj_params3", "--verbosity", "2", "cmd1")
        self.assertEqual(read_django_parameters().get("verbosity", None), 2)

        call_command("dj_params3", ["cmd1"], verbosity=2)
        self.assertEqual(read_django_parameters().get("verbosity", None), 2)

        run_command("dj_params3", "--verbosity", "0", "cmd2")
        self.assertEqual(read_django_parameters().get("verbosity", None), 0)

        call_command("dj_params3", ["cmd2"], verbosity=0)
        self.assertEqual(read_django_parameters().get("verbosity", None), 0)

        run_command("dj_params4")
        self.assertEqual(read_django_parameters().get("verbosity", None), 1)

        call_command("dj_params4")
        self.assertEqual(read_django_parameters().get("verbosity", None), 1)

        run_command("dj_params4", "--verbosity", "2")
        self.assertEqual(read_django_parameters().get("verbosity", None), 2)

        call_command("dj_params4", [], verbosity=2)
        self.assertEqual(read_django_parameters().get("verbosity", None), 2)

        run_command("dj_params4", "--verbosity", "0")
        self.assertEqual(read_django_parameters().get("verbosity", None), 0)

        call_command("dj_params4", [], verbosity=0)
        self.assertEqual(read_django_parameters().get("verbosity", None), 0)


class TestDefaultParamOverrides(TestCase):
    """
    Tests that overloaded group/command names work as expected.
    """

    def test_override_direct(self):
        override = get_command("override")
        self.assertDictEqual(
            override("path/to/settings", version="1.1"),
            {"settings": "path/to/settings", "version": "1.1"},
        )

    def test_override_cli(self):
        from tests.apps.test_app.management.commands.override import (
            VersionEnum,
        )

        result = run_command("override", "path/to/settings", "--version", "1.1")[0]
        self.assertEqual(
            result.strip(),
            str(
                {
                    "settings": Path("path/to/settings"),
                    "version": VersionEnum.VERSION1_1,
                }
            ).strip(),
        )

    def test_override_call_command(self):
        from tests.apps.test_app.management.commands.override import (
            VersionEnum,
        )

        call_command("override", "path/to/settings", 1, version="1.1")
        self.assertDictEqual(
            call_command("override", "path/to/settings", 1, version="1.1"),
            {
                "settings": Path("path/to/settings"),
                "version": VersionEnum.VERSION1_1,
                "optional_arg": 1,
            },
        )

    def test_no_suppression(self):
        self.assertEqual(run_command("dj_params_none_suppressed")[2], 0)
        call_command("dj_params_none_suppressed")

    def test_all_suppressed(self):
        self.assertEqual(run_command("dj_params_all_suppressed")[2], 0)
        call_command("dj_params_all_suppressed")

    def test_all_suppressed_init(self):
        self.assertEqual(run_command("dj_params_all_suppressed_init", "cmd")[2], 0)
        call_command("dj_params_all_suppressed_init", "cmd")

    def test_some_suppressed(self):
        stdout, _, retcode = run_command("dj_params_some_suppressed")
        self.assertEqual(retcode, 0)
        self.assertEqual(stdout.strip(), "traceback=True")

        self.assertEqual(call_command("dj_params_some_suppressed"), "traceback=True")

        stdout, _, retcode = run_command("dj_params_some_suppressed", "--no-traceback")
        self.assertEqual(retcode, 0)
        self.assertEqual(stdout.strip(), "traceback=False")

        self.assertEqual(
            call_command("dj_params_some_suppressed", "--no-traceback"),
            "traceback=False",
        )
        self.assertEqual(
            call_command("dj_params_some_suppressed", traceback=False),
            "traceback=False",
        )

    def test_some_suppressed_init(self, command_name="dj_params_some_suppressed_init"):
        stdout, _, retcode = run_command(command_name)
        self.assertEqual(retcode, 0)
        self.assertTrue("traceback=True" in stdout.strip())

        stdout, _, retcode = run_command(command_name, "cmd")
        self.assertEqual(retcode, 0)
        self.assertTrue("traceback=True" in stdout.strip())

        self.assertTrue("traceback=True" in call_command(command_name))
        self.assertTrue("traceback=True" in call_command(command_name, "cmd"))

        stdout, _, retcode = run_command(command_name, "--no-traceback")
        self.assertEqual(retcode, 0)
        self.assertTrue("traceback=False" in stdout.strip())

        self.assertTrue(
            "traceback=False" in call_command(command_name, "--no-traceback")
        )
        self.assertTrue(
            "traceback=False" in call_command(command_name, traceback=False), "False"
        )

        stdout, _, retcode = run_command(command_name, "--no-traceback", "cmd")
        self.assertEqual(retcode, 0)
        self.assertTrue("traceback=False" in stdout.strip())

        self.assertTrue(
            "traceback=False" in call_command(command_name, "--no-traceback", "cmd")
        )
        self.assertTrue(
            "traceback=False" in call_command(command_name, "cmd", traceback=False)
        )

    def test_some_suppressed_inherit(self):
        self.test_some_suppressed_init("dj_params_inherit")

    def test_some_suppressed_subgroups(self):
        self.test_some_suppressed_init("dj_params_subgroups")
        stdout, _, retcode = run_command(
            "dj_params_subgroups", "--no-traceback", "subgroup", "--no-skip-checks"
        )
        self.assertEqual(retcode, 0)
        self.assertTrue("traceback=False, skipchecks=False" in stdout.strip())

        self.assertTrue(
            "traceback=True, skipchecks=True"
            in call_command(
                "dj_params_subgroups", "--traceback", "subgroup", "--skip-checks"
            )
        )
        self.assertTrue(
            "traceback=False, skipchecks=True"
            in call_command(
                "dj_params_subgroups", "subgroup", traceback=False, skip_checks=True
            )
        )
